from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract, get_event_data

from raiden_contracts.constants import (
    CONTRACT_TOKEN_NETWORK,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    EVENT_DEPRECATION_SWITCH,
    EVENT_TOKEN_NETWORK_CREATED,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
)
from raiden_contracts.contract_manager import ContractManager
from raiden_contracts.tests.fixtures.channel import call_settle
from raiden_contracts.tests.utils import ChannelValues, LockedAmounts, call_and_transact
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.utils.pending_transfers import (
    get_pending_transfers_tree_with_generated_lists,
)


def test_controller(
    web3: Web3,
    contracts_manager: ContractManager,
    deploy_tester_contract: Callable,
    secret_registry_contract: Contract,
    custom_token: Contract,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
    get_accounts: Callable,
) -> None:
    """A creates a TokenNetworkRegistry and B registers a TokenNetwork

    This test is mainly a happy-path scenario. One Ethereum account creates a
    TokenNetworkRegistry, registers a TokenNetwork. TokenNetworkRegistry emits
    events that shows the new TokenNetwork. The new TokenNetwork has the
    original account as the deprecation executor. During these, this test also
    tries to register more than one TokenNetworks and see a failure.
    """
    (controller, B) = get_accounts(2)

    token_network_registry = deploy_tester_contract(
        CONTRACT_TOKEN_NETWORK_REGISTRY,
        controller,
        _secret_registry_address=secret_registry_contract.address,
        _chain_id=web3.eth.chain_id,
        _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
        _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
        _max_token_networks=1,
    )

    # Make sure deployer is controller
    assert token_network_registry.functions.controller().call() == controller
    assert token_network_registry.functions.token_network_created().call() == 0

    # We can only deploy one TokenNetwork contract
    # It can be deployed by anyone
    tx_hash = call_and_transact(
        token_network_registry.functions.createERC20TokenNetwork(
            custom_token.address,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        ),
        {"from": B},
    )
    assert token_network_registry.functions.token_network_created().call() == 1

    # No other TokenNetworks can be deployed now
    with pytest.raises(TransactionFailed, match="TNR: registry full"):
        token_network_registry.functions.createERC20TokenNetwork(
            custom_token.address,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        ).call({"from": B})
    with pytest.raises(TransactionFailed, match="TNR: registry full"):
        token_network_registry.functions.createERC20TokenNetwork(
            custom_token.address,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        ).call({"from": controller})

    tx_receipt = web3.eth.get_transaction_receipt(tx_hash)
    event_abi = contracts_manager.get_event_abi(
        CONTRACT_TOKEN_NETWORK_REGISTRY, EVENT_TOKEN_NETWORK_CREATED
    )
    event_data = get_event_data(web3.codec, event_abi, tx_receipt["logs"][0])
    token_network_address = event_data["args"]["token_network_address"]
    token_network = web3.eth.contract(
        abi=contracts_manager.get_contract_abi(CONTRACT_TOKEN_NETWORK),
        address=token_network_address,
    )

    assert token_network.functions.controller().call() == controller


def test_set_deprecation_switch(
    get_accounts: Callable,
    token_network: Contract,
    web3: Web3,
    contracts_manager: ContractManager,
) -> None:
    """The deprecation executor deprecates a TokenNetwork contract"""
    (A) = get_accounts(1)[0]
    controller = token_network.functions.controller().call()

    assert token_network.functions.safety_deprecation_switch().call() is False

    with pytest.raises(TransactionFailed, match="Can only be called by controller"):
        token_network.functions.deprecate().call({"from": A})

    tx = call_and_transact(token_network.functions.deprecate(), {"from": controller})
    assert token_network.functions.safety_deprecation_switch().call() is True
    tx_receipt = web3.eth.get_transaction_receipt(tx)
    event_abi = contracts_manager.get_event_abi(CONTRACT_TOKEN_NETWORK, EVENT_DEPRECATION_SWITCH)
    event_data = get_event_data(web3.codec, event_abi, tx_receipt["logs"][0])
    assert event_data["args"]["new_value"]

    # We should not be able to call it again
    with pytest.raises(TransactionFailed, match="TN: network is deprecated"):
        token_network.functions.deprecate().call({"from": A})


def test_deprecation_switch(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
) -> None:
    """Test the effects of the deprecation switch on deposits and channel opening"""

    controller = token_network.functions.controller().call()
    (A, B, C, D) = get_accounts(4)
    deposit = 100
    bigger_deposit = 200

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, deposit, B)
    channel_deposit(channel_identifier, B, deposit, A)

    call_and_transact(token_network.functions.deprecate(), {"from": controller})
    assert token_network.functions.safety_deprecation_switch().call() is True

    # Now we cannot deposit in existent channels
    with pytest.raises(TransactionFailed, match="TN: network is deprecated"):
        channel_deposit(channel_identifier, A, bigger_deposit, B)
    with pytest.raises(TransactionFailed, match="TN: network is deprecated"):
        channel_deposit(channel_identifier, B, bigger_deposit, A)

    # Now we cannot open channels anymore
    with pytest.raises(TransactionFailed, match="TN: network is deprecated"):
        channel_identifier = create_channel(C, D)[0]


def test_deprecation_switch_settle(
    web3: Web3,
    get_accounts: Callable,
    token_network: Contract,
    custom_token: Contract,
    reveal_secrets: Callable,
    create_channel: Callable,
    channel_deposit: Callable,
    close_and_update_channel: Callable,
) -> None:
    """Channel close and settlement still work after the depracation switch is turned on"""
    controller = token_network.functions.controller().call()
    (A, B) = get_accounts(2)
    deposit = 100

    (vals_A, vals_B) = (
        ChannelValues(
            deposit=deposit,
            withdrawn=0,
            transferred=5,
            locked_amounts=LockedAmounts(claimable_locked=2, unclaimable_locked=4),
        ),
        ChannelValues(
            deposit=deposit,
            withdrawn=0,
            transferred=10,
            locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=6),
        ),
    )

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, vals_A.deposit, B)
    channel_deposit(channel_identifier, B, vals_B.deposit, A)

    # Mock pending transfers data for A -> B
    pending_transfers_tree_A = get_pending_transfers_tree_with_generated_lists(
        web3,
        unlockable_amount=vals_A.locked_amounts.claimable_locked,
        expired_amount=vals_A.locked_amounts.unclaimable_locked,
    )
    vals_A.locksroot = pending_transfers_tree_A.hash_of_packed_transfers
    # Reveal A's secrets.
    reveal_secrets(A, pending_transfers_tree_A.unlockable)

    # Mock pending transfers data for B -> A
    pending_transfers_tree_B = get_pending_transfers_tree_with_generated_lists(
        web3,
        unlockable_amount=vals_B.locked_amounts.claimable_locked,
        expired_amount=vals_B.locked_amounts.unclaimable_locked,
    )
    vals_B.locksroot = pending_transfers_tree_B.hash_of_packed_transfers
    # Reveal B's secrets
    reveal_secrets(B, pending_transfers_tree_B.unlockable)

    # Set the deprecation switch to true
    call_and_transact(token_network.functions.deprecate(), {"from": controller})
    assert token_network.functions.safety_deprecation_switch().call() is True

    # We need to make sure we can still close, settle & unlock the channels
    close_and_update_channel(channel_identifier, A, vals_A, B, vals_B)
    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)

    call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)

    # Unlock B's pending transfers that were sent to A
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree_B.packed_transfers
        )
    )

    # Unlock A's pending transfers that were sent to B
    call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree_A.packed_transfers
        )
    )

    assert custom_token.functions.balanceOf(A).call() == pre_balance_A + 107
    assert custom_token.functions.balanceOf(B).call() == pre_balance_B + 93
    assert custom_token.functions.balanceOf(token_network.address).call() == pre_balance_contract



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract
from web3.exceptions import ValidationError

from raiden_contracts.constants import EMPTY_ADDRESS, TEST_SETTLE_TIMEOUT_MIN, ChannelEvent
from raiden_contracts.tests.fixtures.channel import call_settle
from raiden_contracts.tests.utils import (
    EMPTY_ADDITIONAL_HASH,
    EMPTY_BALANCE_HASH,
    EMPTY_SIGNATURE,
    NOT_ADDRESS,
    UINT256_MAX,
    ChannelValues,
    call_and_transact,
)
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.utils.events import check_new_deposit


def test_deposit_channel_call(
    token_network: Contract,
    custom_token: Contract,
    create_channel: Callable,
    get_accounts: Callable,
) -> None:
    """Calling setTotalDeposit() fails with various invalid inputs"""
    (A, B) = get_accounts(2)
    deposit_A = 200
    channel_identifier = create_channel(A, B)[0]

    call_and_transact(custom_token.functions.mint(deposit_A), {"from": A})

    call_and_transact(
        custom_token.functions.approve(token_network.address, deposit_A), {"from": A}
    )

    # Validation failure with an invalid channel identifier
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit(-1, A, deposit_A, B)
    # Validation failure with the empty string instead of a channel identifier
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit("", A, deposit_A, B)
    # Validation failure with a negative number instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit(channel_identifier, -1, A, deposit_A)
    # Validation failure with an empty string instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit(channel_identifier, "", deposit_A, B)
    # Validation failure with an odd-length string instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit(channel_identifier, NOT_ADDRESS, deposit_A, B)
    # Validation failure with the number zero instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit(channel_identifier, 0x0, deposit_A, B)
    # Validation failure with the empty string instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit(channel_identifier, A, deposit_A, "")
    # Validation failure with an odd-length string instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit(channel_identifier, A, deposit_A, NOT_ADDRESS)
    # Validation failure with the number zero instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit(channel_identifier, A, deposit_A, 0x0)
    # Validation failure with a negative amount of deposit
    with pytest.raises(ValidationError):
        token_network.functions.setTotalDeposit(channel_identifier, A, -1, B)
    # Transaction failure with the zero address
    with pytest.raises(TransactionFailed, match="TN: participant address zero"):
        token_network.functions.setTotalDeposit(
            channel_identifier, EMPTY_ADDRESS, deposit_A, B
        ).call({"from": A})
    # Transaction failure with the zero address
    with pytest.raises(TransactionFailed, match="TN: partner address zero"):
        token_network.functions.setTotalDeposit(
            channel_identifier, A, deposit_A, EMPTY_ADDRESS
        ).call({"from": A})
    # Transaction failure with zero total deposit
    with pytest.raises(TransactionFailed, match="TN/deposit: total_deposit is zero"):
        token_network.functions.setTotalDeposit(channel_identifier, A, 0, B).call({"from": A})

    call_and_transact(
        token_network.functions.setTotalDeposit(channel_identifier, A, deposit_A, B),
        {"from": A},
    )
    assert (
        deposit_A
        == token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()[0]
    )
    assert (
        0 == token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()[0]
    )


def test_deposit_notapproved(
    token_network: Contract,
    custom_token: Contract,
    create_channel: Callable,
    get_accounts: Callable,
    web3: Web3,
) -> None:
    """Calling setTotalDeposit() fails without approving transfers on the token contract"""
    (A, B) = get_accounts(2)
    channel_identifier = create_channel(A, B)[0]
    deposit_A = 1

    call_and_transact(custom_token.functions.mint(deposit_A), {"from": A})
    mine_blocks(web3, 1)
    balance = custom_token.functions.balanceOf(A).call()
    assert balance >= deposit_A, f"minted {deposit_A} but the balance is still {balance}"

    with pytest.raises(TransactionFailed):  # TODO: why doesn"t this throw an error on transfer?
        token_network.functions.setTotalDeposit(channel_identifier, A, deposit_A, B).call(
            {"from": A}
        )
    assert (
        0 == token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()[0]
    )


def test_null_or_negative_deposit_fail(
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    assign_tokens: Callable,
    get_accounts: Callable,
) -> None:
    """setTotalDeposit() fails when the total deposit does not increase"""
    (A, B) = get_accounts(2)
    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, 2, B)

    assign_tokens(A, 1)

    # setTotalDeposit is idempotent
    with pytest.raises(TransactionFailed, match="TN/deposit: no deposit added"):
        token_network.functions.setTotalDeposit(channel_identifier, A, 2, B).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/deposit: deposit underflow"):
        token_network.functions.setTotalDeposit(channel_identifier, A, 1, B).call({"from": A})
    assert (
        2 == token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()[0]
    )


def test_deposit_delegate_works(
    get_accounts: Callable,
    create_channel: Callable,
    channel_deposit: Callable,
    token_network: Contract,
) -> None:
    """A third party can successfully call setTokenDeposit()"""
    (A, B, C) = get_accounts(3)
    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, 2, B, tx_from=C)
    assert (
        2 == token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()[0]
    )


def test_deposit_wrong_channel(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    assign_tokens: Callable,
) -> None:
    """setTotalDeposit() with a wrong channelID fails"""
    (A, B, C) = get_accounts(3)
    channel_identifier = create_channel(A, B)[0]
    channel_identifier2 = create_channel(A, C)[0]
    assign_tokens(A, 10)

    with pytest.raises(TransactionFailed, match="TN/deposit: channel id mismatch"):
        token_network.functions.setTotalDeposit(channel_identifier2, A, 10, B).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/deposit: channel id mismatch"):
        token_network.functions.setTotalDeposit(channel_identifier, A, 10, C).call({"from": A})

    call_and_transact(
        token_network.functions.setTotalDeposit(channel_identifier, A, 10, B),
        {"from": A},
    )
    assert (
        10 == token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()[0]
    )


@pytest.mark.skip("Not necessary with limited deposits for the test release.")
def test_channel_deposit_overflow(
    get_accounts: Callable, create_channel: Callable, channel_deposit: Callable
) -> None:
    (A, B) = get_accounts(2)
    deposit_A = 50
    deposit_B_ok = UINT256_MAX - deposit_A
    deposit_B_fail = deposit_B_ok + 1

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)

    with pytest.raises(TransactionFailed, match="TN/deposit: deposit overflow"):
        channel_deposit(channel_identifier, B, deposit_B_fail, A)

    channel_deposit(channel_identifier, B, deposit_B_ok, A)


def test_deposit_channel_state(
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
) -> None:
    """Observe how setTotalDeposit() changes the results of getChannelParticipantInfo()"""
    (A, B) = get_accounts(2)
    A_planned_deposit = 10
    B_planned_deposit = 15

    channel_identifier = create_channel(A, B)[0]

    A_onchain_deposit = token_network.functions.getChannelParticipantInfo(
        channel_identifier, A, B
    ).call()[0]
    assert A_onchain_deposit == 0

    B_onchain_deposit = token_network.functions.getChannelParticipantInfo(
        channel_identifier, B, A
    ).call()[0]
    assert B_onchain_deposit == 0

    channel_deposit(channel_identifier, A, A_planned_deposit, B)
    A_onchain_deposit = token_network.functions.getChannelParticipantInfo(
        channel_identifier, A, B
    ).call()[0]
    assert A_onchain_deposit == A_planned_deposit
    B_onchain_deposit = token_network.functions.getChannelParticipantInfo(
        channel_identifier, B, A
    ).call()[0]
    assert B_onchain_deposit == 0

    channel_deposit(channel_identifier, B, B_planned_deposit, A)
    A_onchain_deposit = token_network.functions.getChannelParticipantInfo(
        channel_identifier, A, B
    ).call()[0]
    assert A_onchain_deposit == A_planned_deposit
    B_onchain_deposit = token_network.functions.getChannelParticipantInfo(
        channel_identifier, B, A
    ).call()[0]
    assert B_onchain_deposit == B_planned_deposit


def test_deposit_wrong_state_fail(
    web3: Web3,
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    assign_tokens: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """setTotalDeposit() fails on Closed or Settled channels."""
    (A, B) = get_accounts(2)
    vals_A = ChannelValues(deposit=2, transferred=0)
    vals_B = ChannelValues(deposit=2, transferred=0)
    channel_identifier = create_channel(A, B, TEST_SETTLE_TIMEOUT_MIN)[0]
    assign_tokens(A, vals_A.deposit)
    assign_tokens(B, vals_B.deposit)
    call_and_transact(
        token_network.functions.setTotalDeposit(channel_identifier, A, vals_A.deposit, B),
        {"from": A},
    )
    call_and_transact(
        token_network.functions.setTotalDeposit(channel_identifier, B, vals_B.deposit, A),
        {"from": B},
    )
    assert (
        vals_A.deposit
        == token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()[0]
    )
    assert (
        vals_B.deposit
        == token_network.functions.getChannelParticipantInfo(channel_identifier, B, A).call()[0]
    )

    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)

    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier,
            non_closing_participant=B,
            closing_participant=A,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=closing_sig,
        ),
        {"from": A},
    )

    assign_tokens(A, 10)
    assign_tokens(B, 10)
    vals_A.deposit += 5
    vals_B.deposit += 5
    with pytest.raises(TransactionFailed, match="TN: channel not open"):
        token_network.functions.setTotalDeposit(channel_identifier, A, vals_A.deposit, B).call(
            {"from": A}
        )
    with pytest.raises(TransactionFailed, match="TN: channel not open"):
        token_network.functions.setTotalDeposit(channel_identifier, B, vals_B.deposit, A).call(
            {"from": B}
        )

    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)
    call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)
    with pytest.raises(TransactionFailed, match="TN: channel not open"):
        token_network.functions.setTotalDeposit(channel_identifier, A, vals_A.deposit, B).call(
            {"from": A}
        )
    with pytest.raises(TransactionFailed, match="TN: channel not open"):
        token_network.functions.setTotalDeposit(channel_identifier, B, vals_B.deposit, A).call(
            {"from": B}
        )


def test_deposit_channel_event(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    event_handler: Callable,
) -> None:
    """setTotalDeposit() from each participant causes a DEPOSIT event"""
    ev_handler = event_handler(token_network)
    (A, B) = get_accounts(2)
    deposit_A = 10
    deposit_B = 15

    channel_identifier = create_channel(A, B)[0]

    txn_hash = channel_deposit(channel_identifier, A, deposit_A, B)
    assert (
        deposit_A
        == token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()[0]
    )

    ev_handler.add(
        txn_hash,
        ChannelEvent.DEPOSIT,
        check_new_deposit(channel_identifier, A, deposit_A),
    )

    txn_hash = channel_deposit(channel_identifier, B, deposit_B, A)
    ev_handler.add(
        txn_hash,
        ChannelEvent.DEPOSIT,
        check_new_deposit(channel_identifier, B, deposit_B),
    )

    ev_handler.check()



from typing import Callable, Tuple

import pytest
from eth.constants import ZERO_ADDRESS
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import UserDepositEvent
from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.tests.utils.blockchain import mine_blocks


def test_deposit(
    user_deposit_contract: Contract, custom_token: Contract, get_accounts: Callable[[int], Tuple]
) -> None:
    (A, B) = get_accounts(2)
    call_and_transact(custom_token.functions.mint(100), {"from": A})
    call_and_transact(
        custom_token.functions.approve(user_deposit_contract.address, 30), {"from": A}
    )

    # deposit to A's own balance
    call_and_transact(user_deposit_contract.functions.deposit(A, 10), {"from": A})
    assert user_deposit_contract.functions.balances(A).call() == 10
    assert user_deposit_contract.functions.total_deposit(A).call() == 10
    assert custom_token.functions.balanceOf(A).call() == 90
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 10

    # increase A's deposit
    call_and_transact(user_deposit_contract.functions.deposit(A, 20), {"from": A})
    assert user_deposit_contract.functions.balances(A).call() == 20
    assert user_deposit_contract.functions.total_deposit(A).call() == 20
    assert custom_token.functions.balanceOf(A).call() == 80
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 20

    # a deposit can't be decreased by calling deposit
    with pytest.raises(TransactionFailed, match="deposit not increasing"):
        user_deposit_contract.functions.deposit(A, 19).call({"from": A})

    # A deposits to the benefit of B
    call_and_transact(user_deposit_contract.functions.deposit(B, 10), {"from": A})
    assert user_deposit_contract.functions.balances(B).call() == 10
    assert user_deposit_contract.functions.total_deposit(B).call() == 10
    assert custom_token.functions.balanceOf(A).call() == 70
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 30

    # Can't deposit more than the token contract allows
    with pytest.raises(TransactionFailed):
        user_deposit_contract.functions.deposit(A, 21).call({"from": A})

    # Can't deposit more than the whole_balance_limit
    limit = user_deposit_contract.functions.whole_balance_limit().call()
    assert limit > 0
    call_and_transact(custom_token.functions.mint(limit + 1), {"from": A})
    call_and_transact(
        custom_token.functions.approve(user_deposit_contract.address, limit + 1),
        {"from": A},
    )
    with pytest.raises(TransactionFailed, match="too much deposit"):
        user_deposit_contract.functions.deposit(A, limit + 1).call({"from": A})


def test_transfer(
    uninitialized_user_deposit_contract: Contract,
    udc_transfer_contract: Contract,
    get_accounts: Callable[[int], Tuple],
    event_handler: Callable,
    custom_token: Contract,
) -> None:
    user_deposit_contract = uninitialized_user_deposit_contract
    ev_handler = event_handler(user_deposit_contract)
    (A, B) = get_accounts(2)
    call_and_transact(custom_token.functions.mint(10), {"from": A})
    call_and_transact(
        custom_token.functions.approve(user_deposit_contract.address, 10), {"from": A}
    )
    call_and_transact(user_deposit_contract.functions.deposit(A, 10), {"from": A})

    # only trusted contracts can call transfer (init has not been called, yet)
    with pytest.raises(TransactionFailed, match="unknown caller"):
        udc_transfer_contract.functions.transfer(A, B, 10).call()

    # happy case
    call_and_transact(
        user_deposit_contract.functions.init(
            udc_transfer_contract.address, udc_transfer_contract.address
        )
    )
    tx_hash = call_and_transact(udc_transfer_contract.functions.transfer(A, B, 10))
    ev_handler.assert_event(tx_hash, UserDepositEvent.BALANCE_REDUCED, dict(owner=A, newBalance=0))
    assert user_deposit_contract.functions.balances(A).call() == 0
    assert user_deposit_contract.functions.balances(B).call() == 10

    # no tokens left
    assert not udc_transfer_contract.functions.transfer(A, B, 1).call()

    # (not) enough tokens left
    assert udc_transfer_contract.functions.transfer(B, A, 10).call()
    assert not udc_transfer_contract.functions.transfer(B, A, 11).call()


def test_deposit_after_transfer(
    uninitialized_user_deposit_contract: Contract,
    udc_transfer_contract: Contract,
    custom_token: Contract,
    get_accounts: Callable[[int], Tuple],
) -> None:
    """Make sure that `total_deposit` and `balance` are not mixed up.

    When doing a deposit followed by a transfer, both variables start to differ
    and we can use another deposit to verify that each is handled correctly.
    """
    user_deposit_contract = uninitialized_user_deposit_contract
    call_and_transact(
        user_deposit_contract.functions.init(
            udc_transfer_contract.address, udc_transfer_contract.address
        )
    )
    (A, B) = get_accounts(2)
    call_and_transact(custom_token.functions.mint(100), {"from": A})
    call_and_transact(
        custom_token.functions.approve(user_deposit_contract.address, 30), {"from": A}
    )

    # deposit + transact
    call_and_transact(user_deposit_contract.functions.deposit(A, 10), {"from": A})
    call_and_transact(udc_transfer_contract.functions.transfer(A, B, 10))
    assert user_deposit_contract.functions.balances(A).call() == 0
    assert user_deposit_contract.functions.total_deposit(A).call() == 10
    assert custom_token.functions.balanceOf(A).call() == 90
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 10

    # check after another deposit
    call_and_transact(user_deposit_contract.functions.deposit(A, 20), {"from": A})
    assert user_deposit_contract.functions.balances(A).call() == 10
    assert user_deposit_contract.functions.total_deposit(A).call() == 20
    assert custom_token.functions.balanceOf(A).call() == 80
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 20


def test_withdraw(
    user_deposit_contract: Contract,
    deposit_to_udc: Callable,
    get_accounts: Callable,
    web3: Web3,
    event_handler: Callable,
) -> None:
    """Test the interaction between planWithdraw, withdraw and effectiveBalance"""
    ev_handler = event_handler(user_deposit_contract)
    (A,) = get_accounts(1)
    deposit_to_udc(A, 30)
    assert user_deposit_contract.functions.balances(A).call() == 30
    assert user_deposit_contract.functions.effectiveBalance(A).call() == 30

    # plan withdraw of 20 tokens
    tx_hash = call_and_transact(user_deposit_contract.functions.planWithdraw(20), {"from": A})
    ev_handler.assert_event(
        tx_hash,
        UserDepositEvent.WITHDRAW_PLANNED,
        dict(withdrawer=A, plannedBalance=10),
    )
    assert user_deposit_contract.functions.balances(A).call() == 30
    assert user_deposit_contract.functions.effectiveBalance(A).call() == 10

    # withdraw won't work before withdraw_delay elapsed
    withdraw_delay = user_deposit_contract.functions.withdraw_delay().call()
    mine_blocks(web3, withdraw_delay - 1)
    with pytest.raises(TransactionFailed, match="withdrawing too early"):
        user_deposit_contract.functions.withdraw(18).call({"from": A})

    # can't withdraw more then planned
    mine_blocks(web3, 1)  # now withdraw_delay is over
    with pytest.raises(TransactionFailed, match="withdrawing more than planned"):
        user_deposit_contract.functions.withdraw(21).call({"from": A})

    # actually withdraw 18 tokens
    call_and_transact(user_deposit_contract.functions.withdraw(18), {"from": A})
    assert user_deposit_contract.functions.balances(A).call() == 12
    assert user_deposit_contract.functions.effectiveBalance(A).call() == 12


def test_withdraw_to_beneficiary(
    user_deposit_contract: Contract,
    deposit_to_udc: Callable,
    get_accounts: Callable,
    web3: Web3,
    event_handler: Callable,
    custom_token: Contract,
) -> None:
    """Test the interaction between planWithdraw, withdrawToBeneficiary and effectiveBalance"""
    ev_handler = event_handler(user_deposit_contract)
    (A, B) = get_accounts(2)
    deposit_to_udc(A, 30)
    assert user_deposit_contract.functions.balances(A).call() == 30
    assert user_deposit_contract.functions.effectiveBalance(A).call() == 30

    # plan withdraw of 20 tokens
    tx_hash = call_and_transact(user_deposit_contract.functions.planWithdraw(20), {"from": A})
    ev_handler.assert_event(
        tx_hash,
        UserDepositEvent.WITHDRAW_PLANNED,
        dict(withdrawer=A, plannedBalance=10),
    )
    assert user_deposit_contract.functions.balances(A).call() == 30
    assert user_deposit_contract.functions.effectiveBalance(A).call() == 10

    # beneficiary can not be zero address
    with pytest.raises(TransactionFailed, match="beneficiary is zero"):
        user_deposit_contract.functions.withdrawToBeneficiary(18, ZERO_ADDRESS).call({"from": A})

    # withdraw won't work before withdraw_delay elapsed
    withdraw_delay = user_deposit_contract.functions.withdraw_delay().call()
    mine_blocks(web3, withdraw_delay - 1)
    with pytest.raises(TransactionFailed, match="withdrawing too early"):
        user_deposit_contract.functions.withdrawToBeneficiary(18, B).call({"from": A})

    # can't withdraw more then planned
    mine_blocks(web3, 1)  # now withdraw_delay is over
    with pytest.raises(TransactionFailed, match="withdrawing more than planned"):
        user_deposit_contract.functions.withdrawToBeneficiary(21, B).call({"from": A})

    # actually withdraw 18 tokens
    assert custom_token.functions.balanceOf(B).call() == 0
    call_and_transact(user_deposit_contract.functions.withdrawToBeneficiary(18, B), {"from": A})
    assert user_deposit_contract.functions.balances(A).call() == 12
    assert user_deposit_contract.functions.effectiveBalance(A).call() == 12

    assert custom_token.functions.balanceOf(B).call() == 18



from typing import Optional, Tuple

from hexbytes import HexBytes
from web3 import Web3
from web3._utils.threads import Timeout
from web3.exceptions import TransactionNotFound
from web3.types import TxData, TxReceipt


def check_successful_tx(
    web3: Web3, txid: HexBytes, timeout: int = 180
) -> Tuple[TxReceipt, TxData]:
    """See if transaction went through (Solidity code did not throw).
    :return: Transaction receipt and transaction info
    """
    receipt = wait_for_transaction_receipt(web3=web3, txid=txid, timeout=timeout)
    if receipt is None:
        raise RuntimeError("Could not obtain a transaction receipt.")
    txinfo = web3.eth.get_transaction(txid)
    if "status" not in receipt:
        raise KeyError(
            'A transaction receipt does not contain the "status" field. '
            "Does your chain have Byzantium rules enabled?"
        )
    if receipt["status"] == 0:
        raise ValueError("Status 0 indicates failure")
    if txinfo["gas"] == receipt["gasUsed"]:
        raise ValueError(f'Gas is completely used ({txinfo["gas"]}). Failure?')
    return receipt, txinfo


def wait_for_transaction_receipt(
    web3: Web3, txid: HexBytes, timeout: int = 180
) -> Optional[TxReceipt]:
    receipt = None
    with Timeout(timeout) as time:
        while not receipt or not receipt["blockNumber"]:  # pylint: disable=E1136
            try:
                receipt = web3.eth.get_transaction_receipt(txid)
            except TransactionNotFound:
                time.sleep(5)

    return receipt



from typing import Callable

import pytest
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_SERVICE_REGISTRY, EMPTY_ADDRESS
from raiden_contracts.tests.utils import (
    DEFAULT_BUMP_DENOMINATOR,
    DEFAULT_BUMP_NUMERATOR,
    DEFAULT_DECAY_CONSTANT,
    DEFAULT_MIN_PRICE,
    DEFAULT_REGISTRATION_DURATION,
    DEPLOYER_ADDRESS,
)


@pytest.fixture(scope="session")
def service_registry(deploy_tester_contract: Callable, custom_token: Contract) -> Contract:
    return deploy_tester_contract(
        CONTRACT_SERVICE_REGISTRY,
        _token_for_registration=custom_token.address,
        _controller=DEPLOYER_ADDRESS,
        _initial_price=int(3000e18),
        _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
        _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
        _decay_constant=DEFAULT_DECAY_CONSTANT,
        _min_price=DEFAULT_MIN_PRICE,
        _registration_duration=DEFAULT_REGISTRATION_DURATION,
    )


@pytest.fixture(scope="session")
def service_registry_without_controller(
    deploy_tester_contract: Callable, custom_token: Contract
) -> Contract:
    return deploy_tester_contract(
        CONTRACT_SERVICE_REGISTRY,
        _token_for_registration=custom_token.address,
        _controller=EMPTY_ADDRESS,
        _initial_price=3000 * (10 ** 18),
        _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
        _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
        _decay_constant=DEFAULT_DECAY_CONSTANT,
        _min_price=DEFAULT_MIN_PRICE,
        _registration_duration=DEFAULT_REGISTRATION_DURATION,
    )



import random

from raiden_contracts.tests.utils.constants import UINT256_MAX
from raiden_contracts.utils.signature import private_key_to_address
from raiden_contracts.utils.type_aliases import PrivateKey


def get_random_privkey() -> PrivateKey:
    """Returns a random private key"""
    return PrivateKey(random.randint(1, UINT256_MAX).to_bytes(32, byteorder="big"))


def get_random_address() -> str:
    """Returns a random valid ethereum address"""
    return private_key_to_address(get_random_privkey())



from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, Iterable

import pytest
from py._path.local import LocalPath

from raiden_contracts.constants import (
    ALDERAAN_VERSION,
    BESPIN_VERSION,
    CHAINNAME_TO_ID,
    CONTRACT_TOKEN_NETWORK,
    CONTRACTS_VERSION,
    PRECOMPILED_DATA_FIELDS,
    ChannelEvent,
)
from raiden_contracts.contract_manager import (
    ContractManager,
    ContractManagerLoadError,
    contracts_deployed_path,
    contracts_precompiled_path,
)
from raiden_contracts.contract_source_manager import (
    ContractSourceManager,
    ContractSourceManagerVerificationError,
    contracts_source_path,
    verify_single_precompiled_checksum_on_nonexistent_contract_name,
)


def check_precompiled_content(
    manager: ContractManager, contract_names: Iterable[str], fields: Iterable[str]
) -> None:
    for contract_name in contract_names:
        for field in fields:
            assert manager.contracts[contract_name][field]  # type: ignore


def test_nonexistent_precompiled_path() -> None:
    """An exception occurs when trying to access a field in a non-existent precompiled path"""
    nonexistent_version = "0.6.0"
    with pytest.raises(FileNotFoundError):
        ContractManager(contracts_precompiled_path(nonexistent_version))


def test_verification_overall_checksum() -> None:
    """Tamper with the overall checksum and see failures in verify_precompiled_checksums()"""
    manager = ContractSourceManager(contracts_source_path(contracts_version=None))
    manager.verify_precompiled_checksums(contracts_precompiled_path())

    assert manager.overall_checksum
    original_checksum = manager.overall_checksum

    # We change the source code overall checksum
    manager.overall_checksum += "2"
    # Now the verification should fail
    with pytest.raises(ContractSourceManagerVerificationError):
        manager.verify_precompiled_checksums(contracts_precompiled_path())

    manager.overall_checksum = None  # type: ignore
    with pytest.raises(ContractSourceManagerVerificationError):
        manager.verify_precompiled_checksums(contracts_precompiled_path())

    manager.overall_checksum = ""
    with pytest.raises(ContractSourceManagerVerificationError):
        manager.verify_precompiled_checksums(contracts_precompiled_path())

    checksum_fail = list(original_checksum)
    # Replace the first char with a different one
    checksum_fail[0] = "a" if checksum_fail[0] == "2" else "2"
    manager.overall_checksum = "".join(checksum_fail)
    with pytest.raises(ContractSourceManagerVerificationError):
        manager.verify_precompiled_checksums(contracts_precompiled_path())

    manager.overall_checksum = original_checksum
    manager.verify_precompiled_checksums(contracts_precompiled_path())


def test_verification_contracts_checksums() -> None:
    """Tamper with the contract checksums and see failures in verify_precompiled_checksums()"""
    manager = ContractSourceManager(contracts_source_path(contracts_version=None))
    manager.verify_precompiled_checksums(contracts_precompiled_path())

    assert manager.contracts_checksums
    for contract, checksum in manager.contracts_checksums.items():
        manager.contracts_checksums[contract] += "2"
        with pytest.raises(ContractSourceManagerVerificationError):
            manager.verify_precompiled_checksums(contracts_precompiled_path())

        manager.contracts_checksums[contract] = None  # type: ignore
        with pytest.raises(ContractSourceManagerVerificationError):
            manager.verify_precompiled_checksums(contracts_precompiled_path())

        manager.contracts_checksums[contract] = ""
        with pytest.raises(ContractSourceManagerVerificationError):
            manager.verify_precompiled_checksums(contracts_precompiled_path())

        checksum_fail = list(checksum)
        # Replace the first char with a different one
        checksum_fail[0] = "a" if checksum_fail[0] == "2" else "2"
        manager.contracts_checksums[contract] = "".join(checksum_fail)
        with pytest.raises(ContractSourceManagerVerificationError):
            manager.verify_precompiled_checksums(contracts_precompiled_path())

        manager.contracts_checksums[contract] = checksum
        manager.verify_precompiled_checksums(contracts_precompiled_path())


def test_current_development_version() -> None:
    """contracts_source_path() exists and contains the expected files"""
    contracts_version = CONTRACTS_VERSION
    contract_names = [
        "Utils",
        "SecretRegistry",
        "TokenNetworkRegistry",
        "TokenNetwork",
        "MonitoringService",
        "ServiceRegistry",
    ]

    manager = ContractManager(contracts_precompiled_path(contracts_version))
    assert manager.contracts_version == contracts_version
    check_precompiled_content(manager, contract_names, PRECOMPILED_DATA_FIELDS)

    for _, source_path in contracts_source_path(contracts_version=None).items():
        assert source_path.exists()
    assert contracts_precompiled_path().exists()

    # deployment files exist
    assert contracts_deployed_path(CHAINNAME_TO_ID["rinkeby"]).exists()
    assert contracts_deployed_path(CHAINNAME_TO_ID["ropsten"]).exists()
    # deployment files for service contracts also exist
    assert contracts_deployed_path(CHAINNAME_TO_ID["rinkeby"], services=True).exists()
    assert contracts_deployed_path(CHAINNAME_TO_ID["ropsten"], services=True).exists()


def test_red_eyes_version() -> None:
    """contracts_source_path('0.4.0') exists and contains the expected files"""
    contracts_version = "0.4.0"
    contract_names = [
        "Utils",
        "EndpointRegistry",
        "SecretRegistry",
        "TokenNetworkRegistry",
        "TokenNetwork",
    ]

    manager = ContractManager(contracts_precompiled_path(contracts_version))
    assert manager.contracts_version == contracts_version
    check_precompiled_content(manager, contract_names, PRECOMPILED_DATA_FIELDS)

    assert contracts_precompiled_path(contracts_version).exists()
    assert contracts_deployed_path(CHAINNAME_TO_ID["mainnet"], contracts_version).exists()
    assert contracts_deployed_path(CHAINNAME_TO_ID["rinkeby"], contracts_version).exists()
    assert contracts_deployed_path(CHAINNAME_TO_ID["ropsten"], contracts_version).exists()


def test_alderaan_version() -> None:
    """contracts_source_path('0.37.0') exists and contains the expected files"""
    contract_names = [
        "Utils",
        "SecretRegistry",
        "TokenNetworkRegistry",
        "TokenNetwork",
        "ServiceRegistry",
        "MonitoringService",
        "UserDeposit",
        "OneToN",
    ]

    manager = ContractManager(contracts_precompiled_path(ALDERAAN_VERSION))
    assert manager.contracts_version == ALDERAAN_VERSION
    check_precompiled_content(manager, contract_names, PRECOMPILED_DATA_FIELDS)

    assert contracts_precompiled_path(ALDERAAN_VERSION).exists()
    assert contracts_deployed_path(CHAINNAME_TO_ID["mainnet"], ALDERAAN_VERSION).exists()
    assert contracts_deployed_path(CHAINNAME_TO_ID["rinkeby"], ALDERAAN_VERSION).exists()
    assert contracts_deployed_path(CHAINNAME_TO_ID["ropsten"], ALDERAAN_VERSION).exists()
    assert contracts_deployed_path(CHAINNAME_TO_ID["goerli"], ALDERAAN_VERSION).exists()


def contract_manager_meta(contracts_path: Path) -> None:
    """See failures in looking up non-existent ABI entries of TokenNetwork and CLOSED"""
    manager = ContractManager(contracts_path)

    abi = manager.get_contract_abi(CONTRACT_TOKEN_NETWORK)
    assert isinstance(abi, list)
    with pytest.raises(KeyError):
        manager.get_contract_abi("SomeName")
    with pytest.raises(KeyError):
        manager.get_contract("SomeName")

    abi2 = manager.get_event_abi(CONTRACT_TOKEN_NETWORK, ChannelEvent.CLOSED)
    assert isinstance(abi2, dict)
    with pytest.raises(ValueError):
        manager.get_event_abi(CONTRACT_TOKEN_NETWORK, "NonExistant")


def contract_source_manager_meta(contracts_path: Dict[str, Path]) -> None:
    """See failures in looking up non-existent ABI entries of TokenNetwork and CLOSED"""
    with NamedTemporaryFile() as tmpfile:
        manager = ContractSourceManager(contracts_path).compile_contracts(Path(tmpfile.name))

        abi = manager.get_contract_abi(CONTRACT_TOKEN_NETWORK)
        assert isinstance(abi, list)
        with pytest.raises(KeyError):
            manager.get_contract_abi("SomeName")
        with pytest.raises(KeyError):
            manager.get_contract("SomeName")

        abi2 = manager.get_event_abi(CONTRACT_TOKEN_NETWORK, ChannelEvent.CLOSED)
        assert isinstance(abi2, dict)
        with pytest.raises(ValueError):
            manager.get_event_abi(CONTRACT_TOKEN_NETWORK, "NonExistant")


def test_contract_manager_without_contracts() -> None:
    """ContractManager's constructor fails on a JSON with "contracts": null"""
    with NamedTemporaryFile() as tmpfile:
        tmpfile.write(b'{"contracts": null}')
        tmpfile.flush()
        with pytest.raises(RuntimeError):
            ContractManager(Path(tmpfile.name))


def test_contract_manager_compile() -> None:
    """Check the ABI in the sources"""
    contract_source_manager_meta(contracts_source_path(contracts_version=None))


def test_contract_manager_json(tmpdir: LocalPath) -> None:
    """Check the ABI in contracts.json"""
    precompiled_path = Path(tmpdir).joinpath("contracts.json")
    ContractSourceManager(contracts_source_path(contracts_version=None)).compile_contracts(
        precompiled_path
    )
    # try to load contracts from a precompiled file
    contract_manager_meta(precompiled_path)


def test_contract_manager_constructor_does_not_invent_version() -> None:
    """ContractManager should not invent a version string"""
    manager = ContractManager(contracts_precompiled_path(version=None))
    assert manager.contracts_version is None


@pytest.mark.parametrize("version", [CONTRACTS_VERSION, ALDERAAN_VERSION, BESPIN_VERSION, "0.4.0"])
def test_contract_manager_constructor_keeps_existing_versions(version: str) -> None:
    """ContractManager should keep an existing version string"""
    manager = ContractManager(contracts_precompiled_path(version=version))
    assert manager.contracts_version == version


def test_contract_manager_precompiled_load_error() -> None:
    """ContractManager's constructor raises ContractManagerLoadError when precompiled
    contracts cannot be loaded"""
    with NamedTemporaryFile() as empty_file:
        with pytest.raises(ContractManagerLoadError):
            ContractManager(Path(empty_file.name))


def test_contract_source_manager_constructor_with_wrong_type() -> None:
    """ConstructSourceManager's constructor raises TypeError on a wrong kind of argument"""
    with pytest.raises(TypeError):
        ContractSourceManager(None)  # type: ignore


def test_verify_single_precompiled_cyhecksum_on_nonexistent_contract_name() -> None:
    with pytest.raises(ContractSourceManagerVerificationError, match="No checksum for"):
        verify_single_precompiled_checksum_on_nonexistent_contract_name()



from typing import Any, Dict, List

import requests_mock
from click.testing import CliRunner
from web3.types import ABI

from raiden_contracts.constants import (
    CONTRACT_MONITORING_SERVICE,
    CONTRACT_ONE_TO_N,
    CONTRACT_SECRET_REGISTRY,
    CONTRACT_SERVICE_REGISTRY,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    CONTRACT_USER_DEPOSIT,
    DeploymentModule,
)
from raiden_contracts.contract_manager import (
    CompiledContract,
    ContractManager,
    DeployedContracts,
    contracts_precompiled_path,
)
from raiden_contracts.deploy.etherscan_verify import (
    api_of_chain_id,
    etherscan_verify,
    get_constructor_args,
    guid_status,
    join_sources,
)

contract_name = "DummyContract"


def test_get_constructor_args_no_args() -> None:
    """Test get_constructor_args() on no arguments"""
    contract_manager = ContractManager(contracts_precompiled_path())
    deploy_info: Dict = {"contracts": {contract_name: {"constructor_arguments": []}}}
    assert get_constructor_args(deploy_info, contract_name, contract_manager) == ""  # type: ignore


def abi_with_constructor_input_types(types: List[str]) -> ABI:
    return [{"type": "constructor", "inputs": [{"type": ty} for ty in types]}]  # type: ignore


def test_get_constructor_args_one_arg() -> None:
    """Test get_constructor_args() on one argument"""
    contract_manager = ContractManager(contracts_precompiled_path())
    contract_manager.contracts[contract_name] = CompiledContract(
        {
            "abi": abi_with_constructor_input_types(["uint256"]),
            "bin": "dummy",
            "bin-runtime": "dummy",
            "metadata": "dummy",
        }
    )
    deploy_info: DeployedContracts = {
        "contracts": {contract_name: {"constructor_arguments": [16]}}  # type: ignore
    }
    assert (
        get_constructor_args(deploy_info, contract_name, contract_manager)
        == "0000000000000000000000000000000000000000000000000000000000000010"
    )


def test_get_constructor_args_two_args() -> None:
    """Test get_constructor_args() on two arguments"""
    contract_manager = ContractManager(contracts_precompiled_path())
    contract_manager.contracts[contract_name] = CompiledContract(
        {
            "abi": abi_with_constructor_input_types(["uint256", "bool"]),
            "bin-runtime": "dummy",
            "bin": "dummy",
            "metadata": "dummy",
        }
    )
    deploy_info: DeployedContracts = {
        "contracts": {contract_name: {"constructor_arguments": [16, True]}}  # type: ignore
    }
    assert (
        get_constructor_args(deploy_info, contract_name, contract_manager)
        == "0000000000000000000000000000000000000000000000000000000000000010"
        "0000000000000000000000000000000000000000000000000000000000000001"
    )


def test_run_join_contracts() -> None:
    """Just running join_sources()"""
    join_sources(DeploymentModule.RAIDEN, CONTRACT_TOKEN_NETWORK_REGISTRY)
    join_sources(DeploymentModule.RAIDEN, CONTRACT_SECRET_REGISTRY)
    join_sources(DeploymentModule.SERVICES, CONTRACT_MONITORING_SERVICE)
    join_sources(DeploymentModule.SERVICES, CONTRACT_SERVICE_REGISTRY)
    join_sources(DeploymentModule.SERVICES, CONTRACT_ONE_TO_N)
    join_sources(DeploymentModule.SERVICES, CONTRACT_USER_DEPOSIT)


def test_guid_status() -> None:
    with requests_mock.Mocker() as m:
        etherscan_api = api_of_chain_id[3]
        m.get(etherscan_api, text='{ "content": 1 }')
        assert guid_status(etherscan_api, "something", apikey="foo") == {"content": 1}


def test_etherscan_verify_with_guid() -> None:
    with requests_mock.Mocker() as m:
        chain_id = 3
        etherscan_api = api_of_chain_id[chain_id]
        m.get(etherscan_api, text='{ "content": 1 }')
        runner = CliRunner()
        result = runner.invoke(
            etherscan_verify,
            ["--chain-id", str(chain_id), "--apikey", "API", "--guid", "something"],
        )
        assert result.exit_code == 0


def test_etherscan_verify_already_verified() -> None:
    with requests_mock.Mocker() as m:
        chain_id = 3
        etherscan_api = api_of_chain_id[chain_id]
        m.post(
            etherscan_api,
            text="""
            {
                "status": "0",
                "result" : "Contract source code already verified",
                "message" : ""
            }
            """,
        )
        runner = CliRunner()
        runner.invoke(
            etherscan_verify,
            [
                "--chain-id",
                str(chain_id),
                "--apikey",
                "API",
                "--contract-name",
                "SecretRegistry",
            ],
            catch_exceptions=False,
        )


def test_etherscan_verify_unknown_error() -> None:
    with requests_mock.Mocker() as m:
        chain_id = 3
        etherscan_api = api_of_chain_id[chain_id]
        m.post(
            etherscan_api,
            text="""
            {
                "status": "0",'
                "result" : "Unknown message",
                "message" : ""
            }
            """,
        )
        runner = CliRunner()
        result = runner.invoke(
            etherscan_verify,
            [
                "--chain-id",
                str(chain_id),
                "--apikey",
                "API",
                "--contract-name",
                "SecretRegistry",
            ],
        )
        assert result.exit_code != 0


def test_etherscan_verify_unable_to_verify() -> None:
    with requests_mock.Mocker() as m:
        chain_id = 3
        etherscan_api = api_of_chain_id[chain_id]
        m.post(
            etherscan_api,
            text="""
            {
                "status": "1",
                "result" : "dummy_guid",
                "message" : ""
            }
            """,
        )
        m.get(
            etherscan_api,
            text="""
            {
                "status": "0",
                "result" : "Fail - Unable to verify",
                "message" : ""
            }
            """,
        )
        runner = CliRunner()
        result = runner.invoke(
            etherscan_verify,
            [
                "--chain-id",
                str(chain_id),
                "--apikey",
                "API",
                "--contract-name",
                "SecretRegistry",
            ],
        )
        assert result.exit_code != 0


def test_etherscan_verify_success() -> None:
    with requests_mock.Mocker() as m:
        chain_id = 3
        etherscan_api = api_of_chain_id[chain_id]
        m.post(etherscan_api, text='{ "status": "1", "result" : "guid", "message" : "" }')
        m.get(
            etherscan_api,
            text="""
            {
                "status": "1",
                "result" : "Pass - Verified",
                "message" : ""
            }
            """,
        )
        runner = CliRunner()
        result = runner.invoke(
            etherscan_verify,
            [
                "--chain-id",
                str(chain_id),
                "--apikey",
                "API",
                "--contract-name",
                "SecretRegistry",
            ],
        )
        assert result.exit_code == 0


def first_fail_second_succeed(_: Any, context: Any) -> str:
    """Simulate Etherscan saying for the first time 'wait', but for the second time 'success'."""
    context.status_code = 200
    try:
        if first_fail_second_succeed.called:  # type: ignore
            return '{ "status": "1", "result" : "Pass - Verified", "message" : "" }'
    except AttributeError:  # first time
        pass
    first_fail_second_succeed.called = True  # type: ignore
    return '{ "status": "0", "result" : "wait for a moment", "message" : "" }'


def test_etherscan_verify_success_after_a_loop() -> None:
    with requests_mock.Mocker() as m:
        chain_id = 3
        etherscan_api = api_of_chain_id[chain_id]
        m.post(etherscan_api, text='{ "status": "1", "result" : "guid", "message" : "" }')
        m.get(etherscan_api, text=first_fail_second_succeed)
        runner = CliRunner()
        result = runner.invoke(
            etherscan_verify,
            [
                "--chain-id",
                str(chain_id),
                "--apikey",
                "API",
                "--contract-name",
                "SecretRegistry",
            ],
        )
        assert result.exit_code == 0


def test_etherscan_verify_fail_unknown_contract() -> None:
    """When invoked with an unknown contract name, etherscan_verify should error"""
    chain_id = 3
    runner = CliRunner()
    result = runner.invoke(
        etherscan_verify,
        ["--chain-id", str(chain_id), "--apikey", "API", "--contract-name", "Unknown"],
    )
    assert result.exit_code != 0
    assert result.exception
    assert "unknown contract name" in result.output



from typing import Callable

import pytest
from eth_tester import EthereumTester
from eth_utils import is_address
from eth_utils.units import units
from web3.contract import Contract

from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.tests.utils.constants import FAUCET_ADDRESS


@pytest.fixture
def send_funds(ethereum_tester: EthereumTester, custom_token: Contract) -> Callable:
    """Send some tokens and eth to specified address."""

    def f(target: str) -> None:
        assert is_address(target)
        ethereum_tester.send_transaction(
            {
                "from": FAUCET_ADDRESS,
                "to": target,
                "gas": 21000,
                "value": 1 * int(units["ether"]),
            }
        )
        call_and_transact(
            custom_token.functions.transfer(_to=target, _value=10000),
            {"from": FAUCET_ADDRESS},
        )

    return f



from typing import Callable, Dict

import pytest
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_CUSTOM_TOKEN, CONTRACT_HUMAN_STANDARD_TOKEN

CUSTOM_TOKEN_TOTAL_SUPPLY = 10 ** 26


@pytest.fixture(scope="session")
def token_args() -> Dict:
    return {
        "initial_supply": CUSTOM_TOKEN_TOTAL_SUPPLY,
        "decimal_units": 18,
        "token_name": CONTRACT_CUSTOM_TOKEN,
        "token_symbol": "TKN",
    }


@pytest.fixture(scope="session")
def custom_token_factory(deploy_tester_contract: Callable, token_args: Dict) -> Callable:
    """A function that deploys a CustomToken contract"""

    def f() -> Contract:
        return deploy_tester_contract(CONTRACT_CUSTOM_TOKEN, **token_args)

    return f


@pytest.fixture(scope="session")
def custom_token(custom_token_factory: Callable) -> Contract:
    """Deploy CustomToken contract"""
    return custom_token_factory()


@pytest.fixture(scope="session")
def zero_supply_custom_token(deploy_tester_contract: Callable) -> Contract:
    """Deploy a CustomToken contract with zero initial supply"""
    return deploy_tester_contract(CONTRACT_CUSTOM_TOKEN, (0, 18, "ZeroToken", "ZRO"))


@pytest.fixture()
def human_standard_token(deploy_token_contract: Callable, token_args: Dict) -> Contract:
    """Deploy HumanStandardToken contract"""
    return deploy_token_contract(
        _initialAmount=token_args["initial_supply"],
        _decimalUnits=token_args["decimal_units"],
        _tokenName=token_args["token_name"],
        _tokenSymbol=token_args["token_symbol"],
    )


@pytest.fixture
def deploy_token_contract(deploy_tester_contract: Callable) -> Callable:
    """Returns a function that deploys a generic HumanStandardToken contract"""

    def f(**args: Dict) -> Contract:
        return deploy_tester_contract(CONTRACT_HUMAN_STANDARD_TOKEN, **args)

    return f


@pytest.fixture
def standard_token_contract(custom_token: Contract) -> Contract:
    """Deployed CustomToken contract"""
    return custom_token



from typing import Callable, Dict, List, Optional

import pytest
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import (
    CONTRACT_MONITORING_SERVICE,
    CONTRACT_ONE_TO_N,
    CONTRACT_SECRET_REGISTRY,
    CONTRACT_SERVICE_REGISTRY,
    CONTRACT_TOKEN_NETWORK,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    CONTRACT_USER_DEPOSIT,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
    MessageTypeId,
)
from raiden_contracts.contract_manager import gas_measurements
from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.tests.utils.constants import DEPLOYER_ADDRESS, SERVICE_DEPOSIT, UINT256_MAX
from raiden_contracts.utils.pending_transfers import get_locked_amount, get_pending_transfers_tree
from raiden_contracts.utils.proofs import sign_one_to_n_iou, sign_reward_proof
from raiden_contracts.utils.type_aliases import BlockExpiration, ChainID, TokenAmount


@pytest.mark.parametrize("version", [None])
def test_gas_json_has_enough_fields(version: Optional[str]) -> None:
    """Check is gas.json contains enough fields"""
    doc = gas_measurements(version)
    keys = {
        "CustomToken.mint",
        "CustomToken.approve",
        "CustomToken.transfer",
        "CustomToken.transferFrom",
        "MonitoringService.claimReward",
        "MonitoringService.monitor",
        "OneToN.claim",
        "OneToN.bulkClaim 1 ious",
        "OneToN.bulkClaim 6 ious",
        "SecretRegistry.registerSecret",
        "SecretRegistry.registerSecretBatch1",
        "SecretRegistry.registerSecretBatch2",
        "SecretRegistry.registerSecretBatch3",
        "ServiceRegistry.deposit",
        "ServiceRegistry.setURL",
        "TokenNetwork DEPLOYMENT",
        "TokenNetwork.closeChannel",
        "TokenNetwork.cooperativeSettle",
        "TokenNetwork.openChannel",
        "TokenNetwork.openChannelWithDeposit",
        "TokenNetwork.setTotalDeposit",
        "TokenNetwork.setTotalWithdraw",
        "TokenNetwork.settleChannel",
        "TokenNetwork.unlock 1 locks",
        "TokenNetwork.unlock 6 locks",
        "TokenNetwork.updateNonClosingBalanceProof",
        "TokenNetworkRegistry DEPLOYMENT",
        "TokenNetworkRegistry createERC20TokenNetwork",
        "UserDeposit.deposit",
        "UserDeposit.deposit (increase balance)",
        "UserDeposit.planWithdraw",
        "UserDeposit.withdraw",
    }
    assert set(doc.keys()) == keys


@pytest.fixture
def print_gas_token_network_registry(
    web3: Web3,
    deploy_tester_contract_txhash: Callable,
    secret_registry_contract: Contract,
    print_gas: Callable,
) -> None:
    """Abusing pytest to print the deployment gas cost of TokenNetworkRegistry"""
    txhash = deploy_tester_contract_txhash(
        CONTRACT_TOKEN_NETWORK_REGISTRY,
        _secret_registry_address=secret_registry_contract.address,
        _chain_id=web3.eth.chain_id,
        _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
        _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
        _max_token_networks=10,
    )
    print_gas(txhash, CONTRACT_TOKEN_NETWORK_REGISTRY + " DEPLOYMENT")


@pytest.fixture
def print_gas_token_network_deployment(
    web3: Web3,
    get_accounts: Callable,
    print_gas: Callable,
    custom_token: Contract,
    secret_registry_contract: Contract,
    deploy_tester_contract_txhash: Callable,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
) -> None:
    """Abusing pytest to print the deployment gas cost of TokenNetwork"""
    controller = get_accounts(1)[0]
    txhash = deploy_tester_contract_txhash(
        CONTRACT_TOKEN_NETWORK,
        _token_address=custom_token.address,
        _secret_registry=secret_registry_contract.address,
        _chain_id=web3.eth.chain_id,
        _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
        _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
        _controller=controller,
        _channel_participant_deposit_limit=channel_participant_deposit_limit,
        _token_network_deposit_limit=token_network_deposit_limit,
    )
    print_gas(txhash, CONTRACT_TOKEN_NETWORK + " DEPLOYMENT")


@pytest.fixture
def print_gas_token_network_create(
    print_gas: Callable,
    custom_token: Contract,
    get_token_network_registry: Callable,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
    token_network_registry_constructor_args: Dict,
) -> None:
    """Abusing pytest to print gas cost of TokenNetworkRegistry's createERC20TokenNetwork()"""
    registry = get_token_network_registry(**token_network_registry_constructor_args)
    txn_hash = call_and_transact(
        registry.functions.createERC20TokenNetwork(
            custom_token.address,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        ),
        {"from": DEPLOYER_ADDRESS},
    )

    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK_REGISTRY + " createERC20TokenNetwork")


@pytest.fixture
def print_gas_secret_registry(secret_registry_contract: Contract, print_gas: Callable) -> None:
    """Abusing pytest to print gas cost of SecretRegistry's registerSecret()"""
    secret = b"secretsecretsecretsecretsecretse"
    txn_hash = call_and_transact(secret_registry_contract.functions.registerSecret(secret))
    print_gas(txn_hash, CONTRACT_SECRET_REGISTRY + ".registerSecret")

    for batch_size in (1, 2, 3):
        batch = [bytes([batch_size, i] + [0] * 30) for i in range(batch_size)]
        txn_hash = call_and_transact(secret_registry_contract.functions.registerSecretBatch(batch))
        print_gas(txn_hash, CONTRACT_SECRET_REGISTRY + ".registerSecretBatch" + str(batch_size))


@pytest.fixture
def print_gas_channel_cycle(
    web3: Web3,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    withdraw_channel: Callable,
    secret_registry_contract: Contract,
    get_accounts: Callable,
    print_gas: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """Abusing pytest to print gas costs of TokenNetwork's operations"""
    (A, B, C, D) = get_accounts(4)
    settle_timeout = 11

    (channel_identifier, txn_hash) = create_channel(A, B, settle_timeout)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".openChannel")

    (_, txn_hash) = create_channel(C, D, settle_timeout)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".openChannel")

    txn_hash = channel_deposit(channel_identifier, A, 20, B)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".setTotalDeposit")

    txn_hash = channel_deposit(channel_identifier, B, 10, A)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".setTotalDeposit")

    txn_hash = withdraw_channel(channel_identifier, A, 5, UINT256_MAX, B)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".setTotalWithdraw")

    pending_transfers_tree1 = get_pending_transfers_tree(web3, [1, 1, 2, 3], [2, 1])
    locksroot1 = pending_transfers_tree1.hash_of_packed_transfers
    locked_amount1 = get_locked_amount(pending_transfers_tree1.transfers)

    pending_transfers_tree2 = get_pending_transfers_tree(web3, [3], [], 7)
    locksroot2 = pending_transfers_tree2.hash_of_packed_transfers
    locked_amount2 = get_locked_amount(pending_transfers_tree2.transfers)

    balance_proof_A = create_balance_proof(
        channel_identifier, A, 10, locked_amount1, 5, locksroot1
    )
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )
    balance_proof_B = create_balance_proof(channel_identifier, B, 5, locked_amount2, 3, locksroot2)
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )

    for lock in pending_transfers_tree1.unlockable:
        txn_hash = call_and_transact(
            secret_registry_contract.functions.registerSecret(lock[3]), {"from": A}
        )
    print_gas(txn_hash, CONTRACT_SECRET_REGISTRY + ".registerSecret")

    for lock in pending_transfers_tree2.unlockable:
        txn_hash = call_and_transact(
            secret_registry_contract.functions.registerSecret(lock[3]), {"from": A}
        )
    print_gas(txn_hash, CONTRACT_SECRET_REGISTRY + ".registerSecret")

    txn_hash = call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_B._asdict().values(), closing_sig_A
        ),
        {"from": A},
    )
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".closeChannel")

    txn_hash = call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ),
        {"from": B},
    )
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".updateNonClosingBalanceProof")

    mine_blocks(web3, settle_timeout)
    txn_hash = call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier,
            B,
            5,
            locked_amount2,
            locksroot2,
            A,
            10,
            locked_amount1,
            locksroot1,
        )
    )
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".settleChannel")

    txn_hash = call_and_transact(
        token_network.functions.unlock(
            channel_identifier, A, B, pending_transfers_tree2.packed_transfers
        )
    )
    print_gas(
        txn_hash,
        "{0}.unlock {1} locks".format(
            CONTRACT_TOKEN_NETWORK, len(pending_transfers_tree2.transfers)
        ),
    )

    txn_hash = call_and_transact(
        token_network.functions.unlock(
            channel_identifier, B, A, pending_transfers_tree1.packed_transfers
        )
    )
    print_gas(
        txn_hash,
        "{0}.unlock {1} locks".format(
            CONTRACT_TOKEN_NETWORK, len(pending_transfers_tree1.transfers)
        ),
    )


@pytest.fixture
def print_gas_channel_open_with_deposit(
    token_network: Contract,
    get_accounts: Callable,
    print_gas: Callable,
    assign_tokens: Callable,
) -> None:
    """Abusing pytest to print gas costs of TokenNetwork's `openChannelWithDeposit`"""
    (A, B) = get_accounts(2)
    settle_timeout = 11
    deposit = 20

    assign_tokens(A, deposit)
    tx_hash = call_and_transact(
        token_network.functions.openChannelWithDeposit(A, B, settle_timeout, deposit),
        {"from": A},
    )
    print_gas(tx_hash, CONTRACT_TOKEN_NETWORK + ".openChannelWithDeposit")


@pytest.fixture
def print_gas_coop_settle(
    create_channel: Callable,
    channel_deposit: Callable,
    cooperative_settle_channel: Callable,
    get_accounts: Callable,
    print_gas: Callable,
) -> None:
    """Abusing pytest to print gas costs of cooperative settle"""
    (A, B) = get_accounts(2)
    settle_timeout = 11

    (channel_identifier, txn_hash) = create_channel(A, B, settle_timeout)
    channel_deposit(channel_identifier, A, 20, B)

    txn_hash = cooperative_settle_channel(channel_identifier, A, B, 10, 10, UINT256_MAX, A)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".cooperativeSettle")


@pytest.fixture
def print_gas_monitoring_service(
    token_network: Contract,
    monitoring_service_external: Contract,
    get_accounts: Callable,
    create_channel: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    service_registry: Contract,
    custom_token: Contract,
    deposit_to_udc: Callable,
    print_gas: Callable,
    get_private_key: Callable,
    create_service_account: Callable,
    web3: Web3,
) -> None:
    """Abusing pytest to print gas cost of MonitoringService functions"""
    # setup: two parties + MS
    (A, MS) = get_accounts(2)
    B = create_service_account()
    reward_amount = TokenAmount(10)
    deposit_to_udc(B, reward_amount)

    # register MS in the ServiceRegistry contract
    call_and_transact(custom_token.functions.mint(SERVICE_DEPOSIT * 2), {"from": MS})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT),
        {"from": MS},
    )
    call_and_transact(service_registry.functions.deposit(SERVICE_DEPOSIT), {"from": MS})

    # open a channel (c1, c2)
    channel_identifier = create_channel(A, B)[0]

    # create balance and reward proofs
    balance_proof_A = create_balance_proof(channel_identifier, B, transferred_amount=10, nonce=1)
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_A._asdict(),
    )
    balance_proof_B = create_balance_proof(channel_identifier, A, transferred_amount=20, nonce=2)
    non_closing_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_B._asdict(),
    )
    reward_proof_signature = sign_reward_proof(
        privatekey=get_private_key(B),
        monitoring_service_contract_address=monitoring_service_external.address,
        chain_id=token_network.functions.chain_id().call(),
        token_network_address=token_network.address,
        non_closing_participant=B,
        reward_amount=reward_amount,
        non_closing_signature=non_closing_signature_B,
    )

    # c1 closes channel
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_A._asdict().values(), closing_sig_A
        ),
        {"from": A},
    )
    mine_blocks(web3, 4)

    # MS calls `MSC::monitor()` using c1's BP and reward proof
    txn_hash = call_and_transact(
        monitoring_service_external.functions.monitor(
            A,
            B,
            balance_proof_B.balance_hash,
            balance_proof_B.nonce,
            balance_proof_B.additional_hash,
            balance_proof_B.original_signature,
            non_closing_signature_B,  # non-closing signature
            reward_amount,
            token_network.address,  # token network address
            reward_proof_signature,
        ),
        {"from": MS},
    )
    print_gas(txn_hash, CONTRACT_MONITORING_SERVICE + ".monitor")

    mine_blocks(web3, 1)

    # MS claims the reward
    txn_hash = call_and_transact(
        monitoring_service_external.functions.claimReward(
            channel_identifier, token_network.address, A, B
        ),
        {"from": MS},
    )
    print_gas(txn_hash, CONTRACT_MONITORING_SERVICE + ".claimReward")


@pytest.fixture
def print_gas_one_to_n(
    one_to_n_contract: Contract,
    deposit_to_udc: Callable,
    print_gas: Callable,
    make_iou: Callable,
    web3: Web3,
    get_private_key: Callable,
    create_service_account: Callable,
    create_account: Callable,
) -> None:
    """Abusing pytest to print gas cost of OneToN functions"""
    A = create_account()
    B = create_service_account()
    deposit_to_udc(A, 30)

    # happy case
    chain_id = web3.eth.chain_id
    amount = TokenAmount(10)
    expiration = BlockExpiration(web3.eth.block_number + 2)
    signature = sign_one_to_n_iou(
        get_private_key(A),
        sender=A,
        receiver=B,
        amount=amount,
        expiration_block=expiration,
        one_to_n_address=one_to_n_contract.address,
        chain_id=ChainID(chain_id),
    )
    txn_hash = call_and_transact(
        one_to_n_contract.functions.claim(A, B, amount, expiration, signature),
        {"from": A},
    )

    print_gas(txn_hash, CONTRACT_ONE_TO_N + ".claim")

    # bulk claims gas prices
    def concat_iou_data(ious: List[Dict], key: str) -> List:
        return [iou[key] for iou in ious]

    def concat_iou_signatures(ious: List[Dict]) -> bytes:
        result = b""
        for iou in ious:
            result += iou["signature"]

        return result

    for num_ious in (1, 6):
        receivers = [create_service_account() for i in range(num_ious)]
        ious = [make_iou(A, r) for r in receivers]

        txn_hash = call_and_transact(
            one_to_n_contract.functions.bulkClaim(
                concat_iou_data(ious, "sender"),
                concat_iou_data(ious, "receiver"),
                concat_iou_data(ious, "amount"),
                concat_iou_data(ious, "expiration_block"),
                concat_iou_signatures(ious),
            ),
            {"from": A},
        )
        print_gas(txn_hash, CONTRACT_ONE_TO_N + f".bulkClaim {num_ious} ious")


@pytest.fixture
def print_gas_user_deposit(
    user_deposit_contract: Contract,
    custom_token: Contract,
    get_accounts: Callable,
    web3: Web3,
    print_gas: Callable,
) -> None:
    """Abusing pytest to print gas cost of UserDeposit functions

    The `transfer` function is not included because it's only called by trusted
    contracts as part of another function.
    """
    (A,) = get_accounts(1)
    call_and_transact(custom_token.functions.mint(20), {"from": A})
    call_and_transact(
        custom_token.functions.approve(user_deposit_contract.address, 20), {"from": A}
    )

    # deposit
    txn_hash = call_and_transact(user_deposit_contract.functions.deposit(A, 10), {"from": A})
    print_gas(txn_hash, CONTRACT_USER_DEPOSIT + ".deposit")
    txn_hash = call_and_transact(user_deposit_contract.functions.deposit(A, 20), {"from": A})
    print_gas(txn_hash, CONTRACT_USER_DEPOSIT + ".deposit (increase balance)")

    # plan withdraw
    txn_hash = call_and_transact(user_deposit_contract.functions.planWithdraw(10), {"from": A})
    print_gas(txn_hash, CONTRACT_USER_DEPOSIT + ".planWithdraw")

    # withdraw
    withdraw_delay = user_deposit_contract.functions.withdraw_delay().call()
    mine_blocks(web3, withdraw_delay)
    txn_hash = call_and_transact(user_deposit_contract.functions.withdraw(10), {"from": A})
    print_gas(txn_hash, CONTRACT_USER_DEPOSIT + ".withdraw")


@pytest.fixture
def print_gas_service_registry(
    custom_token: Contract,
    service_registry: Contract,
    print_gas: Callable,
    create_account: Callable,
) -> None:
    A = create_account()
    deposit = service_registry.functions.currentPrice().call()
    call_and_transact(custom_token.functions.mint(deposit), {"from": A})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, deposit), {"from": A}
    )
    deposit_tx = call_and_transact(service_registry.functions.deposit(deposit), {"from": A})
    print_gas(deposit_tx, CONTRACT_SERVICE_REGISTRY + ".deposit")
    url = "http://example.com"
    set_url_tx = call_and_transact(service_registry.functions.setURL(url), {"from": A})
    print_gas(set_url_tx, CONTRACT_SERVICE_REGISTRY + ".setURL")


@pytest.fixture
def print_gas_token(get_accounts: Callable, custom_token: Contract, print_gas: Callable) -> None:
    (A, B) = get_accounts(2)
    tx_hash = call_and_transact(custom_token.functions.mint(100), {"from": A})
    print_gas(tx_hash, "CustomToken.mint")
    tx_hash = call_and_transact(custom_token.functions.transfer(B, 100), {"from": A})
    print_gas(tx_hash, "CustomToken.transfer")
    tx_hash = call_and_transact(custom_token.functions.approve(A, 100), {"from": B})
    print_gas(tx_hash, "CustomToken.approve")
    tx_hash = call_and_transact(custom_token.functions.transferFrom(B, A, 100), {"from": A})
    print_gas(tx_hash, "CustomToken.transferFrom")


# All gas printing is done in a single test. Otherwise, after a parallel
# execution of multiple gas printing tests, you see a corrupted gas.json.
@pytest.mark.slow
@pytest.mark.usefixtures(
    "print_gas_token_network_registry",
    "print_gas_token_network_deployment",
    "print_gas_token_network_create",
    "print_gas_secret_registry",
    "print_gas_channel_cycle",
    "print_gas_channel_open_with_deposit",
    "print_gas_coop_settle",
    "print_gas_monitoring_service",
    "print_gas_one_to_n",
    "print_gas_service_registry",
    "print_gas_user_deposit",
    "print_gas_token",
)
def test_print_gas() -> None:
    pass



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract, get_event_data
from web3.exceptions import MismatchedABI

from raiden_contracts.constants import (
    CONTRACT_DEPOSIT,
    CONTRACT_SERVICE_REGISTRY,
    EMPTY_ADDRESS,
    EVENT_REGISTERED_SERVICE,
)
from raiden_contracts.contract_manager import ContractManager, contracts_precompiled_path
from raiden_contracts.tests.utils import HexBytes, call_and_transact
from raiden_contracts.tests.utils.constants import (
    DEFAULT_BUMP_DENOMINATOR,
    DEFAULT_BUMP_NUMERATOR,
    DEFAULT_DECAY_CONSTANT,
    DEFAULT_MIN_PRICE,
    DEFAULT_REGISTRATION_DURATION,
    DEPLOYER_ADDRESS,
    SECONDS_PER_DAY,
    SERVICE_DEPOSIT,
    UINT256_MAX,
)


def test_deposit_contract(
    get_deposit_contract: Callable, custom_token: Contract, get_accounts: Callable
) -> None:
    """Deposit contract with zero-deadline should release the deposit immediately"""
    (A,) = get_accounts(1)
    call_and_transact(custom_token.functions.mint(100), {"from": A})
    depo = get_deposit_contract(
        _token=custom_token.address, _release_at=0, _withdrawer=A, _service_registry=A
    )
    call_and_transact(custom_token.functions.transfer(depo.address, 100), {"from": A})
    assert custom_token.functions.balanceOf(A).call() == 0
    assert custom_token.functions.balanceOf(depo.address).call() == 100
    call_and_transact(depo.functions.withdraw(A), {"from": A})
    assert custom_token.functions.balanceOf(A).call() == 100
    assert custom_token.functions.balanceOf(depo.address).call() == 0


def test_deposit_contract_without_service_registry_code(
    get_deposit_contract: Callable, custom_token: Contract, get_accounts: Callable
) -> None:
    """If Deposit has no code in service registry, too early withdrawals fail"""
    (A,) = get_accounts(1)
    call_and_transact(custom_token.functions.mint(100), {"from": A})
    depo = get_deposit_contract(
        _token=custom_token.address,
        _release_at=UINT256_MAX,
        _withdrawer=A,
        _service_registry=A,
    )
    call_and_transact(custom_token.functions.transfer(depo.address, 100), {"from": A})
    assert custom_token.functions.balanceOf(A).call() == 0
    assert custom_token.functions.balanceOf(depo.address).call() == 100
    # The call fails because an empty account doesn't return a boolean.
    # In this case, the message of the TransactionFailed exception is ''.
    with pytest.raises(TransactionFailed) as ex:
        call_and_transact(depo.functions.withdraw(A), {"from": A})
    assert str(ex.value) == "execution reverted: b''"

    assert custom_token.functions.balanceOf(A).call() == 0
    assert custom_token.functions.balanceOf(depo.address).call() == 100


def test_deposit_contract_too_early_withdraw(
    get_deposit_contract: Callable,
    custom_token: Contract,
    get_accounts: Callable,
    service_registry: Contract,
) -> None:
    """Deposit contract with some deadline should not release the deposit immediately"""
    (A,) = get_accounts(1)
    call_and_transact(custom_token.functions.mint(100), {"from": A})
    depo = get_deposit_contract(
        _token=custom_token.address,
        _release_at=UINT256_MAX,
        _withdrawer=A,
        _service_registry=service_registry.address,
    )
    call_and_transact(custom_token.functions.transfer(depo.address, 100), {"from": A})
    assert custom_token.functions.balanceOf(A).call() == 0
    assert custom_token.functions.balanceOf(depo.address).call() == 100
    with pytest.raises(TransactionFailed, match="deposit not released yet"):
        call_and_transact(depo.functions.withdraw(A), {"from": A})
    assert custom_token.functions.balanceOf(A).call() == 0
    assert custom_token.functions.balanceOf(depo.address).call() == 100


def test_deposit(
    service_registry: Contract, custom_token: Contract, get_accounts: Callable
) -> None:
    """A service provider can make deposits to ServiceRegistry"""
    (A,) = get_accounts(1)
    call_and_transact(custom_token.functions.mint(SERVICE_DEPOSIT), {"from": A})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT),
        {"from": A},
    )

    # happy path
    old_balance = custom_token.functions.balanceOf(A).call()
    old_price = service_registry.functions.currentPrice().call()
    old_len = service_registry.functions.everMadeDepositsLen().call()
    assert not service_registry.functions.hasValidRegistration(A).call()
    call_and_transact(service_registry.functions.deposit(SERVICE_DEPOSIT), {"from": A})
    assert old_balance > custom_token.functions.balanceOf(A).call() > old_balance - old_price
    assert service_registry.functions.currentPrice().call() > old_price
    assert service_registry.functions.everMadeDepositsLen().call() == old_len + 1
    assert service_registry.functions.hasValidRegistration(A).call()
    first_expiration = service_registry.functions.service_valid_till(A).call()

    # custom_token does not allow transfer of more tokens
    with pytest.raises(TransactionFailed, match="not enough limit"):
        service_registry.functions.deposit(1).call({"from": A})

    # More minting and approving before extending the registration
    call_and_transact(custom_token.functions.mint(SERVICE_DEPOSIT), {"from": A})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT),
        {"from": A},
    )

    # Extending the registration
    call_and_transact(service_registry.functions.deposit(SERVICE_DEPOSIT), {"from": A})
    assert service_registry.functions.hasValidRegistration(A).call()
    second_expiration = service_registry.functions.service_valid_till(A).call()
    assert second_expiration == first_expiration + DEFAULT_REGISTRATION_DURATION
    # This time, the list of addresses that have ever made deposits should not grow
    assert service_registry.functions.everMadeDepositsLen().call() == old_len + 1


def test_setURL(
    custom_token: Contract, service_registry: Contract, get_accounts: Callable, web3: Web3
) -> None:
    """A ServiceRegistry allows registered service providers to set their URLs"""
    (A,) = get_accounts(1)
    url1 = "http://example.com"
    url2 = "http://raiden.example.com"

    call_and_transact(custom_token.functions.mint(SERVICE_DEPOSIT), {"from": A})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT),
        {"from": A},
    )
    tx = call_and_transact(service_registry.functions.deposit(SERVICE_DEPOSIT), {"from": A})
    tx_receipt = web3.eth.get_transaction_receipt(tx)
    contract_manager = ContractManager(contracts_precompiled_path(version=None))
    event_abi = contract_manager.get_event_abi(CONTRACT_SERVICE_REGISTRY, EVENT_REGISTERED_SERVICE)
    event_data = get_event_data(web3.codec, event_abi, tx_receipt["logs"][-1])
    assert event_data["args"]["service"] == A
    assert event_data["args"]["deposit_contract"] != EMPTY_ADDRESS

    call_and_transact(service_registry.functions.setURL(url1), {"from": A})
    assert service_registry.functions.urls(A).call() == url1

    call_and_transact(service_registry.functions.setURL(url2), {"from": A})
    assert service_registry.functions.urls(A).call() == url2


def test_decayed_price(service_registry: Contract) -> None:
    """Test the exponential decayedPrice() function"""
    assert service_registry.functions.decayedPrice(100000, 0).call() == 100000

    # The minimum price is 1000
    assert service_registry.functions.decayedPrice(100, 0).call() == 1000

    # roughly 139 days till the price halves.
    assert service_registry.functions.decayedPrice(100000, 11990300).call() == 50000


def test_changing_duration(
    service_registry: Contract, get_accounts: Callable, custom_token: Contract
) -> None:
    """The controller can change the registration period of ServiceRegistry"""
    new_duration = 90 * SECONDS_PER_DAY
    call_and_transact(
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=new_duration,
        ),
        {"from": DEPLOYER_ADDRESS},
    )
    # make sure that the duration has changed.
    assert service_registry.functions.registration_duration().call() == new_duration
    (A,) = get_accounts(1)
    call_and_transact(custom_token.functions.mint(2 * SERVICE_DEPOSIT), {"from": A})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, 2 * SERVICE_DEPOSIT),
        {"from": A},
    )
    call_and_transact(service_registry.functions.deposit(SERVICE_DEPOSIT), {"from": A})
    first_expiration = service_registry.functions.service_valid_till(A).call()
    call_and_transact(service_registry.functions.deposit(SERVICE_DEPOSIT), {"from": A})
    second_expiration = service_registry.functions.service_valid_till(A).call()
    assert second_expiration == first_expiration + new_duration


def test_changing_duration_to_huge_value(
    service_registry: Contract, get_accounts: Callable, custom_token: Contract
) -> None:
    """When the duration is huge and the deadline overflows, deposit fails"""
    new_duration = 2 ** 256 - 1
    call_and_transact(
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=new_duration,
        ),
        {"from": DEPLOYER_ADDRESS},
    )
    # make sure that the duration has changed.
    assert service_registry.functions.registration_duration().call() == new_duration
    (A,) = get_accounts(1)
    call_and_transact(custom_token.functions.mint(2 * SERVICE_DEPOSIT), {"from": A})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, 2 * SERVICE_DEPOSIT),
        {"from": A},
    )
    with pytest.raises(TransactionFailed, match="overflow during extending the registration"):
        call_and_transact(service_registry.functions.deposit(SERVICE_DEPOSIT), {"from": A})


def test_changing_bump_numerator(service_registry: Contract) -> None:
    """The controller can change the price bump numerator"""
    call_and_transact(
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR + 1,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ),
        {"from": DEPLOYER_ADDRESS},
    )
    assert service_registry.functions.price_bump_numerator().call() == DEFAULT_BUMP_NUMERATOR + 1


def test_calling_internal_bump_paramter_change(service_registry: Contract) -> None:
    """Calling an internal function setPriceBumpParameters() must fail"""
    with pytest.raises(MismatchedABI):
        call_and_transact(
            service_registry.functions.setPriceBumpParameters(
                DEFAULT_BUMP_NUMERATOR + 1, DEFAULT_BUMP_DENOMINATOR
            ),
            {"from": DEPLOYER_ADDRESS},
        )


def test_too_high_bump_numerator_fail(service_registry: Contract) -> None:
    """changeParameters() fails if the numerator is too big"""
    with pytest.raises(TransactionFailed, match="price dump numerator is too big"):
        call_and_transact(
            service_registry.functions.changeParameters(
                _price_bump_numerator=2 ** 40,
                _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
                _decay_constant=DEFAULT_DECAY_CONSTANT,
                _min_price=DEFAULT_MIN_PRICE,
                _registration_duration=DEFAULT_REGISTRATION_DURATION,
            ),
            {"from": DEPLOYER_ADDRESS},
        )


def test_changing_bump_denominator(service_registry: Contract) -> None:
    """The controller can change the price dump denominator"""
    call_and_transact(
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR + 1,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ),
        {"from": DEPLOYER_ADDRESS},
    )
    assert (
        service_registry.functions.price_bump_denominator().call() == DEFAULT_BUMP_DENOMINATOR + 1
    )


def test_changing_too_low_bump_parameter_fail(service_registry: Contract) -> None:
    """changeParameters() fails if the bump numerator is smaller than the bump denominator"""
    with pytest.raises(TransactionFailed, match="price dump instead of bump"):
        call_and_transact(
            service_registry.functions.changeParameters(
                _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
                _price_bump_denominator=DEFAULT_BUMP_NUMERATOR + 1,
                _decay_constant=DEFAULT_DECAY_CONSTANT,
                _min_price=DEFAULT_MIN_PRICE,
                _registration_duration=DEFAULT_REGISTRATION_DURATION,
            ),
            {"from": DEPLOYER_ADDRESS},
        )


def test_zero_numerator_fail(service_registry: Contract) -> None:
    """changeParameters() fails if the bump numerator is set to zero"""
    with pytest.raises(TransactionFailed, match="price dump instead of bump"):
        call_and_transact(
            service_registry.functions.changeParameters(
                _price_bump_numerator=0,
                _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
                _decay_constant=DEFAULT_DECAY_CONSTANT,
                _min_price=DEFAULT_MIN_PRICE,
                _registration_duration=DEFAULT_REGISTRATION_DURATION,
            ),
            {"from": DEPLOYER_ADDRESS},
        )


def test_changing_decay_constant(service_registry: Contract) -> None:
    """The controller can change the price decay constant"""
    call_and_transact(
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=DEFAULT_DECAY_CONSTANT + 100,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ),
        {"from": DEPLOYER_ADDRESS},
    )
    assert service_registry.functions.decay_constant().call() == DEFAULT_DECAY_CONSTANT + 100


def test_very_small_decay_cosntant(service_registry: Contract) -> None:
    """set a very small decay constant and see very fast price decay"""
    call_and_transact(
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=1,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ),
        {"from": DEPLOYER_ADDRESS},
    )
    assert service_registry.functions.decayedPrice(100000, 100).call() == DEFAULT_MIN_PRICE


def test_internal_set_decay_constant(service_registry: Contract) -> None:
    """Calling the internal setDecayConstant() must fail"""
    with pytest.raises(MismatchedABI):
        call_and_transact(
            service_registry.functions.setDecayConstant(DEFAULT_DECAY_CONSTANT + 100),
            {"from": DEPLOYER_ADDRESS},
        )


def test_too_high_decay_constant_fail(service_registry: Contract) -> None:
    """changeParameters() fails if the new decay constant is too high"""
    with pytest.raises(TransactionFailed, match="too big decay constant"):
        call_and_transact(
            service_registry.functions.changeParameters(
                _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
                _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
                _decay_constant=2 ** 40,
                _min_price=DEFAULT_MIN_PRICE,
                _registration_duration=DEFAULT_REGISTRATION_DURATION,
            ),
            {"from": DEPLOYER_ADDRESS},
        )


def test_very_big_decay_cosntant(service_registry: Contract) -> None:
    """set a very big decay constant and see very slow price decay"""
    call_and_transact(
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=2 ** 40 - 1,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ),
        {"from": DEPLOYER_ADDRESS},
    )
    assert service_registry.functions.decayedPrice(100000, 11990300).call() == 99998


def test_zero_denominator_fail(service_registry: Contract) -> None:
    """changeParameters() fails if the new bump denominator is zero"""
    with pytest.raises(TransactionFailed, match="divide by zero"):
        call_and_transact(
            service_registry.functions.changeParameters(
                _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
                _price_bump_denominator=0,
                _decay_constant=DEFAULT_DECAY_CONSTANT,
                _min_price=DEFAULT_MIN_PRICE,
                _registration_duration=DEFAULT_REGISTRATION_DURATION,
            ),
            {"from": DEPLOYER_ADDRESS},
        )


def test_changing_min_price(service_registry: Contract) -> None:
    """The controller can change the min_price"""
    call_and_transact(
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE * 2,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ),
        {"from": DEPLOYER_ADDRESS},
    )
    assert service_registry.functions.min_price().call() == DEFAULT_MIN_PRICE * 2


def test_changing_min_price_above_current(service_registry: Contract) -> None:
    """Changing min_price above the current price."""
    current_price = service_registry.functions.currentPrice().call()
    call_and_transact(
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=current_price + 1,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ),
        {"from": DEPLOYER_ADDRESS},
    )
    assert service_registry.functions.currentPrice().call() == current_price + 1


def test_internal_min_price(service_registry: Contract) -> None:
    """Calling the internal setMinPrice() must fail"""
    with pytest.raises(MismatchedABI):
        call_and_transact(
            service_registry.functions.setMinPrice(DEFAULT_MIN_PRICE * 2),
            {"from": DEPLOYER_ADDRESS},
        )


def test_unauthorized_parameter_change(service_registry: Contract, get_accounts: Callable) -> None:
    """A random address's changeParameters() call should fail"""
    (A,) = get_accounts(1)
    with pytest.raises(TransactionFailed, match="Can only be called by controller"):
        call_and_transact(
            service_registry.functions.changeParameters(
                _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
                _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
                _decay_constant=DEFAULT_DECAY_CONSTANT,
                _min_price=DEFAULT_MIN_PRICE * 2,
                _registration_duration=DEFAULT_REGISTRATION_DURATION,
            ),
            {"from": A},
        )
    assert service_registry.functions.min_price().call() == DEFAULT_MIN_PRICE


def test_parameter_change_on_no_controller(service_registry_without_controller: Contract) -> None:
    """A random address's changeParameters() call should fail"""
    with pytest.raises(TransactionFailed, match="Can only be called by controller"):
        call_and_transact(
            service_registry_without_controller.functions.changeParameters(
                _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
                _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
                _decay_constant=DEFAULT_DECAY_CONSTANT,
                _min_price=DEFAULT_MIN_PRICE * 2,
                _registration_duration=DEFAULT_REGISTRATION_DURATION,
            ),
            {"from": DEPLOYER_ADDRESS},
        )
    assert service_registry_without_controller.functions.min_price().call() == DEFAULT_MIN_PRICE


def service_registry_with_zero_supply_token(
    deploy_tester_contract: Callable, zero_supply_custom_token: Contract
) -> None:
    """Deployment of ServiceRegistry should fail with a token with zero supply"""
    with pytest.raises(TransactionFailed, match="total supply zero"):
        deploy_tester_contract(
            CONTRACT_SERVICE_REGISTRY,
            [
                zero_supply_custom_token.address,
                EMPTY_ADDRESS,
                2 ** 90,
                2 ** 40 - 1,
                1,
                2 ** 40 - 1,
                DEFAULT_MIN_PRICE,
                DEFAULT_REGISTRATION_DURATION,
            ],
        )


# FIXME: can be removed?
def service_registry_with_high_numbers(
    deploy_tester_contract: Callable, custom_token: Contract
) -> None:
    """See if the computation overflows with the highest allowed numbers"""
    contract = deploy_tester_contract(
        CONTRACT_SERVICE_REGISTRY,
        [
            custom_token.address,
            EMPTY_ADDRESS,
            2 ** 90,
            2 ** 40 - 1,
            1,
            2 ** 40 - 1,
            DEFAULT_MIN_PRICE,
            DEFAULT_REGISTRATION_DURATION,
        ],
    )
    assert contract.functions.currentPrice() == 2 ** 90


def test_deprecation_switch(
    service_registry: Contract, get_accounts: Callable, custom_token: Contract
) -> None:
    """The controller turns on the deprecation switch and somebody tries to deposit"""
    # The controller turns on the deprecation switch
    assert not service_registry.functions.deprecated().call()
    call_and_transact(
        service_registry.functions.setDeprecationSwitch(), {"from": DEPLOYER_ADDRESS}
    )
    assert service_registry.functions.deprecated().call()
    # A user tries to make a deposit
    (A,) = get_accounts(1)
    minted_amount = service_registry.functions.currentPrice().call()
    call_and_transact(custom_token.functions.mint(minted_amount), {"from": A})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, minted_amount),
        {"from": A},
    )
    with pytest.raises(TransactionFailed, match="this contract was deprecated"):
        call_and_transact(service_registry.functions.deposit(minted_amount), {"from": A})


def test_deprecation_immediate_payout(
    create_account: Callable, custom_token: Contract, service_registry: Contract, web3: Web3
) -> None:
    """When the deprecation switch is on, deposits can be withdrawn immediately."""
    # A user makes a deposit
    A = create_account()
    minted = service_registry.functions.currentPrice().call()
    call_and_transact(custom_token.functions.mint(minted), {"from": A})
    call_and_transact(
        custom_token.functions.approve(service_registry.address, minted), {"from": A}
    )
    deposit_tx = call_and_transact(service_registry.functions.deposit(minted), {"from": A})
    # The user obtains the deposit address
    deposit_tx_receipt = web3.eth.get_transaction_receipt(deposit_tx)
    contract_manager = ContractManager(contracts_precompiled_path(version=None))
    event_abi = contract_manager.get_event_abi(CONTRACT_SERVICE_REGISTRY, EVENT_REGISTERED_SERVICE)
    event_data = get_event_data(web3.codec, event_abi, deposit_tx_receipt["logs"][-1])
    deposit_address = event_data["args"]["deposit_contract"]
    # And obtains the Deposit contract instance
    deposit_abi = contract_manager.get_contract_abi(CONTRACT_DEPOSIT)
    deposit = web3.eth.contract(abi=deposit_abi, address=deposit_address)
    # The controller turns on the deprecation switch
    call_and_transact(
        service_registry.functions.setDeprecationSwitch(), {"from": DEPLOYER_ADDRESS}
    )
    # The user successfully withdraws the deposit
    call_and_transact(deposit.functions.withdraw(A), {"from": A})
    # The user has all the balance it has minted
    assert minted == custom_token.functions.balanceOf(A).call()
    # The Deposit contract has destroyed itself
    assert web3.eth.get_code(deposit.address) == HexBytes("0x")


def test_unauthorized_deprecation_switch(
    service_registry: Contract, get_accounts: Callable
) -> None:
    """A random account cannot turn on the deprecation switch"""
    (A,) = get_accounts(1)
    with pytest.raises(TransactionFailed, match="Can only be called by controller"):
        call_and_transact(service_registry.functions.setDeprecationSwitch(), {"from": A})


def test_deploying_service_registry_with_denominator_zero(
    deploy_tester_contract: Callable, custom_token: Contract
) -> None:
    """ServiceRegistry's constructor must fail when denominator is zero"""
    # Web3 does not expose the error message "divide by zero"
    # when require() fails in the constructor.
    with pytest.raises(TransactionFailed, match="divide by zero"):
        deploy_tester_contract(
            CONTRACT_SERVICE_REGISTRY,
            _token_for_registration=custom_token.address,
            _controller=DEPLOYER_ADDRESS,
            _initial_price=int(3000e18),
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=0,  # instead of DEFAULT_BUMP_DENOMINATOR
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        )



import string
from random import choice


def fake_hex(size: int, fill: str = "00") -> str:
    return "0x" + "".join([fill for i in range(0, size)])


def fake_bytes(size: int, fill: str = "00") -> bytes:
    return bytes.fromhex(fake_hex(size, fill)[2:])


def make_address() -> bytes:
    return bytes("".join(choice(string.printable) for _ in range(20)), encoding="utf-8")



import functools
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

import click
import requests
from eth_typing import URI, ChecksumAddress
from eth_utils import to_checksum_address
from hexbytes import HexBytes
from web3 import HTTPProvider, Web3
from web3.middleware import construct_sign_and_send_raw_middleware, geth_poa_middleware
from web3.types import TxReceipt, Wei

from raiden_contracts.constants import CONTRACT_CUSTOM_TOKEN
from raiden_contracts.contract_manager import ContractManager, contracts_precompiled_path
from raiden_contracts.utils.private_key import get_private_key
from raiden_contracts.utils.signature import private_key_to_address
from raiden_contracts.utils.transaction import check_successful_tx


class TokenOperations:
    def __init__(
        self, rpc_url: URI, private_key: Path, password: Optional[Path] = None, wait: int = 10
    ):
        self.web3 = Web3(HTTPProvider(rpc_url))
        self.private_key = get_private_key(private_key, password)
        assert self.private_key is not None
        self.owner = private_key_to_address(self.private_key)
        self.wait = wait
        self.web3.middleware_onion.add(construct_sign_and_send_raw_middleware(self.private_key))
        self.web3.eth.default_account = self.owner

        self.web3.middleware_onion.inject(geth_poa_middleware, layer=0)

    def is_valid_contract(self, token_address: ChecksumAddress) -> bool:
        return self.web3.eth.get_code(token_address, "latest") != HexBytes("")

    def mint_tokens(self, token_address: ChecksumAddress, amount: int) -> TxReceipt:
        token_address = to_checksum_address(token_address)
        assert self.is_valid_contract(
            token_address
        ), "The custom token contract does not seem to exist on this address"
        token_contract = ContractManager(contracts_precompiled_path()).get_contract(
            CONTRACT_CUSTOM_TOKEN
        )
        token_proxy = self.web3.eth.contract(address=token_address, abi=token_contract["abi"])
        txhash = token_proxy.functions.mint(amount).transact({"from": self.owner})
        receipt, _ = check_successful_tx(web3=self.web3, txid=txhash, timeout=self.wait)
        return receipt

    def get_weth(self, token_address: ChecksumAddress, amount: int) -> TxReceipt:
        token_address = to_checksum_address(token_address)
        assert (
            self.web3.eth.get_balance(self.owner) > amount
        ), "Not sufficient ether to make a deposit to WETH contract"
        assert self.is_valid_contract(
            token_address
        ), "The WETH token does not exist on this contract"
        result = requests.get(
            "http://api.etherscan.io/api?module=contract&action=getabi&"
            "address=0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
        )
        weth_abi = result.json()["result"]
        weth_proxy = self.web3.eth.contract(address=token_address, abi=weth_abi)
        assert weth_proxy.functions.symbol().call() == "WETH", "This contract is not a WETH token"
        txhash = weth_proxy.functions.deposit().transact(
            {"from": self.owner, "value": Wei(amount)}
        )
        receipt, _ = check_successful_tx(web3=self.web3, txid=txhash, timeout=self.wait)
        return receipt

    # Could be used for both custom token as well as WETH contracts
    def transfer_tokens(self, token_address: ChecksumAddress, dest: str, amount: int) -> TxReceipt:
        token_address = to_checksum_address(token_address)
        dest = to_checksum_address(dest)
        assert self.is_valid_contract(
            token_address
        ), "The token contract does not seem to exist on this address"
        token_contract = ContractManager(contracts_precompiled_path()).get_contract(
            CONTRACT_CUSTOM_TOKEN
        )
        token_proxy = self.web3.eth.contract(address=token_address, abi=token_contract["abi"])
        assert (
            token_proxy.functions.balanceOf(self.owner).call() >= amount
        ), "Not enough token balances"
        txhash = token_proxy.functions.transfer(dest, amount).transact({"from": self.owner})
        receipt, _ = check_successful_tx(web3=self.web3, txid=txhash, timeout=self.wait)
        return receipt

    def get_balance(self, token_address: ChecksumAddress, address: str) -> int:
        token_address = to_checksum_address(token_address)
        address = to_checksum_address(address)
        assert self.is_valid_contract(
            token_address
        ), "The Token Contract does not seem to exist on this address"
        token_contract = ContractManager(contracts_precompiled_path()).get_contract(
            CONTRACT_CUSTOM_TOKEN
        )
        token_proxy = self.web3.eth.contract(address=token_address, abi=token_contract["abi"])
        return token_proxy.functions.balanceOf(address).call()


def common_options(func: Callable) -> Callable:
    """A decorator that combines commonly appearing @click.option decorators."""

    @click.option(
        "--private-key", required=True, help="Path to a private key store.", type=click.STRING
    )
    @click.option("--password", help="password file for the keystore json file", type=click.STRING)
    @click.option(
        "--rpc-url",
        default="http://127.0.0.1:8545",
        help="Address of the Ethereum RPC provider",
        type=click.STRING,
    )
    @click.option(
        "--token-address", required=True, help="Address of the token contract", type=click.STRING
    )
    @click.option(
        "--amount", required=True, help="Amount to mint/deposit/transfer", type=click.INT
    )
    @click.option("--wait", default=300, help="Max tx wait time in s.", type=click.INT)
    @functools.wraps(func)
    def wrapper(*args: List, **kwargs: Dict) -> Any:
        return func(*args, **kwargs)

    return wrapper


@click.group()
def cli() -> None:
    pass


@cli.command()
@common_options
def mint(
    private_key: str,
    password: str,
    rpc_url: URI,
    token_address: ChecksumAddress,
    amount: int,
    wait: int,
) -> None:
    password_file = Path(password) if password else None
    token_ops = TokenOperations(rpc_url, Path(private_key), password_file, wait)
    receipt = token_ops.mint_tokens(token_address, amount)
    print(f"Minting tokens for {token_ops.owner}")
    print(receipt)
    balance = token_ops.get_balance(token_address, token_ops.owner)
    print(f"Balance of the {token_ops.owner} :  {balance}")


@cli.command()
@common_options
def weth(
    private_key: str,
    password: Optional[str],
    rpc_url: URI,
    token_address: ChecksumAddress,
    amount: int,
    wait: int,
) -> None:
    password_path = Path(password) if password else None
    token_ops = TokenOperations(rpc_url, Path(private_key), password_path, wait)
    receipt = token_ops.get_weth(token_address, amount)
    print(f"Getting WETH tokens for {token_ops.owner}")
    print(receipt)
    balance = token_ops.get_balance(token_address, token_ops.owner)
    print(f"Balance of the {token_ops.owner} : {balance}")


@cli.command()
@common_options
@click.option("--destination", help="Address of payee account", type=click.STRING)
def transfer(
    private_key: str,
    password: str,
    rpc_url: URI,
    token_address: ChecksumAddress,
    amount: int,
    wait: int,
    destination: str,
) -> None:
    password_path = Path(password) if password else None
    token_ops = TokenOperations(rpc_url, Path(private_key), password_path, wait)
    receipt = token_ops.transfer_tokens(token_address, destination, amount)
    print(f"Transferring tokens to {destination}")
    print(receipt)
    balance = token_ops.get_balance(token_address, token_ops.owner)
    print(f"Balance of the {token_ops.owner} : {balance}")


@cli.command()
@click.option(
    "--rpc-url",
    default="http://127.0.0.1:8545",
    help="Address of the Ethereum RPC provider",
    type=click.STRING,
)
@click.option(
    "--token-address",
    required=True,
    help="Address of the token contract",
    type=click.STRING,
)
@click.option("--address", help="Address of account to get Balance", type=click.STRING)
def balance(rpc_url: URI, token_address: str, address: str) -> None:
    token_address = to_checksum_address(token_address)
    address = to_checksum_address(address)
    web3 = Web3(HTTPProvider(rpc_url))
    token_contract = ContractManager(contracts_precompiled_path()).get_contract(
        CONTRACT_CUSTOM_TOKEN
    )
    token_proxy = web3.eth.contract(address=token_address, abi=token_contract["abi"])
    balance = token_proxy.functions.balanceOf(address).call()
    print(f"Balance of the {address} : {balance}")


if __name__ == "__main__":
    cli()



import pytest
from eth_typing import BlockNumber
from eth_utils.address import to_checksum_address
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.utils.logs import LogFilter


def test_logfilter_with_nonexistent_event(web3: Web3) -> None:
    """Try to create a LogFilter with a nonexistent event"""

    with pytest.raises(ValueError):
        LogFilter(
            web3=web3,
            abi=[],
            address=to_checksum_address("0xfake"),
            event_name="ev0",
            from_block=BlockNumber(0),
            to_block="latest",
        )


def test_call_and_transact_does_not_mine(web3: Web3, custom_token: Contract) -> None:
    """See call_and_transact() does not mine a block"""

    before = web3.eth.block_number
    call_and_transact(custom_token.functions.multiplier())
    after = web3.eth.block_number
    assert before + 1 == after



from raiden_contracts.tests.utils import UINT256_MAX, ChannelValues, LockedAmounts

# We must cover the edge cases documented in
# https://github.com/raiden-network/raiden-contracts/issues/188
# The scope is to make sure that if someone uses an old balance proof, this cannot be used as
# an attack to steal tokens.
# For invalid balance proofs (created and signed with an unofficial Raiden client),
# we cannot determine and guarantee corectness. There are specific constraints that the
# Raiden client must enforce that guarantee correctness.

# For all valid last balance proofs provided we have manually constructed a suit of old balance
# proof pairs that could be used for attacks (documented edge cases).
# We will test that using old balance proofs does not result in cheating, therefore the attacker
# (participant who provides the old balance proof) must not receive more tokens than in the case
# where he provides a `valid last` balance proof.
channel_settle_test_values = [
    {
        "valid_last": (
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=20020,
                locked_amounts=LockedAmounts(claimable_locked=3, unclaimable_locked=1),
            ),
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=20030,
                locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=2),
            ),
        ),
        # participant2 provides a valid but old balance proof of participant1
        "old_last": [
            # participant2 does not send participant1's balance proof
            ChannelValues(deposit=35, withdrawn=5, transferred=0, locked_amounts=LockedAmounts()),
            # participant2 provides an old participant1 balance proof with a smaller
            # transferred amount
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=10000,
                locked_amounts=LockedAmounts(claimable_locked=3, unclaimable_locked=1),
            ),
            # participant2 provides an old participant1 balance proof with a smaller
            # claimable locked amount
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=20020,
                locked_amounts=LockedAmounts(claimable_locked=0, unclaimable_locked=1),
            ),
            # participant2 provides an old participant1 balance proof with a smaller
            # unclaimable locked amount
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=20020,
                locked_amounts=LockedAmounts(claimable_locked=3, unclaimable_locked=0),
            ),
            # participant2 provides an old participant1 balance proof with a smaller transferred
            # & claimable locked amount
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=10000,
                locked_amounts=LockedAmounts(claimable_locked=0, unclaimable_locked=1),
            ),
            # participant2 provides an old participant1 balance proof will all values smaller
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=10000,
                locked_amounts=LockedAmounts(claimable_locked=0, unclaimable_locked=0),
            ),
            # participant2 provides an old participant1 balance proof, but with the same
            # transferred + claimable_locked
            # This should have the same final participant balances as the valid last balance proofs
            # locked amount cannot be bigger than the available deposit at that time
            # 18 is the maximum locked amount that can happen with valid balance proofs
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=20006,
                locked_amounts=LockedAmounts(claimable_locked=17, unclaimable_locked=1),
            ),
            # participant2 provides an old participant1 balance proof, with a higher
            # unclaimable locked amount can happen if expired transfers are removed
            # from the pending locks
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=20020,
                locked_amounts=LockedAmounts(claimable_locked=3, unclaimable_locked=12),
            ),
            # participant2 provides an old participant1 balance proof with a higher
            # claimable locked amount, but lower transferred + claimable_locked
            # A higher claimable locked amount can happen even if the locked tokens are
            # eventually claimed off-chain (become transferred amount).
            # This is because we can register secrets on-chain at any point in time.
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=10020,
                locked_amounts=LockedAmounts(claimable_locked=17, unclaimable_locked=1),
            ),
            # participant2 provides an old participant1 balance proof with a higher
            # unclaimable locked amount and a higher claimable locked amount but lower
            # transferred + claimable_locked
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=10020,
                locked_amounts=LockedAmounts(claimable_locked=17, unclaimable_locked=10),
            ),
        ],
        # participant1 provides a valid but old participant2 balance proof
        # these examples must maintain the same order of calculating the balances
        # imposed by tranferred2 + locked2 >= transferred1 + locked1
        "last_old": [
            # participant1 provides an old participant2 balance proof with a smaller
            # transferred amount
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=20020,
                locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=2),
            ),
            # participant1 provides an old participant2 balance proof with a smaller
            # claimable locked amount
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=20030,
                locked_amounts=LockedAmounts(claimable_locked=0, unclaimable_locked=2),
            ),
            # participant1 provides an old participant2 balance proof with a smaller
            # unclaimable locked amount
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=20030,
                locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=0),
            ),
            # participant1 provides an old participant2 balance proof with a smaller transferred
            # & claimable locked amount
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=20022,
                locked_amounts=LockedAmounts(claimable_locked=0, unclaimable_locked=2),
            ),
            # participant1 provides an old participant2 balance proof will all values smaller
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=20024,
                locked_amounts=LockedAmounts(claimable_locked=0, unclaimable_locked=0),
            ),
            # participant1 provides an old participant2 balance proof, but with the same
            # transferred + claimable_locked
            # This should have the same final participant balances as the valid last balance proofs
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=19994,
                locked_amounts=LockedAmounts(claimable_locked=40, unclaimable_locked=2),
            ),
            # participant1 provides an old participant2 balance proof, with a higher
            # unclaimable locked amount
            # can happen if expired transfers are removed from the pending locks
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=20030,
                locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=20),
            ),
            # participant1 provides an old participant2 balance proof with a higher
            # claimable locked amount,
            # but lower transferred + claimable_locked
            # A higher claimable locked amount can happen even if the locked tokens are
            # eventually claimed off-chain (become transferred amount).
            # This is because we can register secrets on-chain at any point in time.
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=19990,
                locked_amounts=LockedAmounts(claimable_locked=40, unclaimable_locked=2),
            ),
            # participant1 provides an old participant2 balance proof with a higher
            # unclaimable locked amount and a higher claimable locked amount but lower
            # transferred + claimable_locked
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=19990,
                locked_amounts=LockedAmounts(claimable_locked=40, unclaimable_locked=10),
            ),
        ],
    },
    {
        # neither participants provide balance proofs
        "valid_last": (
            ChannelValues(deposit=40, withdrawn=10, transferred=0),
            ChannelValues(deposit=35, withdrawn=5, transferred=0),
        )
    },
    {
        # both balance proofs provided are valid
        "valid_last": (
            ChannelValues(
                deposit=35,
                withdrawn=5,
                transferred=20,
                locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=0),
            ),
            ChannelValues(
                deposit=40,
                withdrawn=10,
                transferred=30,
                locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=2),
            ),
        )
    },
    {
        # Participants have withdrawn all their tokens already
        "valid_last": (
            ChannelValues(deposit=5, withdrawn=15, transferred=20),
            ChannelValues(deposit=20, withdrawn=10, transferred=30),
        )
    },
    {
        # Participants have withdrawn all their finalized transfer tokens except locked,
        "valid_last": (
            ChannelValues(
                deposit=5,
                withdrawn=5,
                transferred=20,
                locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=1),
            ),
            ChannelValues(
                deposit=25,
                withdrawn=5,
                transferred=30,
                locked_amounts=LockedAmounts(claimable_locked=2, unclaimable_locked=3),
            ),
        )
    },
]

channel_settle_invalid_test_values = [
    (
        # bigger locked amounts than what remains in the contract after settlement
        ChannelValues(
            deposit=35,
            withdrawn=5,
            transferred=20020,
            locked_amounts=LockedAmounts(claimable_locked=30000000, unclaimable_locked=10000000),
        ),
        ChannelValues(
            deposit=40,
            withdrawn=10,
            transferred=20030,
            locked_amounts=LockedAmounts(claimable_locked=10000000, unclaimable_locked=40000000),
        ),
    ),
    # participant2 does not provide a balance proof + locked amount too big
    (
        ChannelValues(
            deposit=35,
            withdrawn=5,
            transferred=0,
            locked_amounts=LockedAmounts(claimable_locked=0, unclaimable_locked=0),
        ),
        ChannelValues(
            deposit=40,
            withdrawn=10,
            transferred=30,
            locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=2),
        ),
    ),
    # Participants have withdrawn all their finalized transfer tokens already,
    # only locked tokens left
    (
        ChannelValues(
            deposit=5,
            withdrawn=10,
            transferred=20,
            locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=1),
        ),
        ChannelValues(
            deposit=20,
            withdrawn=5,
            transferred=30,
            locked_amounts=LockedAmounts(claimable_locked=2, unclaimable_locked=3),
        ),
    ),
    (
        ChannelValues(deposit=5, withdrawn=5, transferred=20),
        ChannelValues(deposit=10, withdrawn=10, transferred=30),
    ),
    (
        ChannelValues(
            deposit=5,
            withdrawn=5,
            transferred=20,
            locked_amounts=LockedAmounts(claimable_locked=1, unclaimable_locked=3),
        ),
        ChannelValues(
            deposit=10,
            withdrawn=10,
            transferred=30,
            locked_amounts=LockedAmounts(claimable_locked=2, unclaimable_locked=4),
        ),
    ),
    # overflow on transferred amounts
    (
        ChannelValues(
            deposit=35,
            withdrawn=5,
            transferred=UINT256_MAX - 15,
            locked_amounts=LockedAmounts(claimable_locked=3, unclaimable_locked=1),
        ),
        ChannelValues(
            deposit=40,
            withdrawn=10,
            transferred=UINT256_MAX - 5,
            locked_amounts=LockedAmounts(claimable_locked=5, unclaimable_locked=1),
        ),
    ),
    # overflow on transferred amount
    (
        ChannelValues(
            deposit=35,
            withdrawn=5,
            transferred=0,
            locked_amounts=LockedAmounts(claimable_locked=4, unclaimable_locked=0),
        ),
        ChannelValues(
            deposit=40,
            withdrawn=10,
            transferred=UINT256_MAX - 5,
            locked_amounts=LockedAmounts(claimable_locked=0, unclaimable_locked=6),
        ),
    ),
    # overflow on transferred amount
    (
        ChannelValues(
            deposit=40,
            withdrawn=10,
            transferred=0,
            locked_amounts=LockedAmounts(claimable_locked=6, unclaimable_locked=0),
        ),
        ChannelValues(
            deposit=35,
            withdrawn=5,
            transferred=UINT256_MAX - 15,
            locked_amounts=LockedAmounts(claimable_locked=1, unclaimable_locked=3),
        ),
    ),
    # overflow on transferred amount
    (
        ChannelValues(
            deposit=35,
            withdrawn=5,
            transferred=20020,
            locked_amounts=LockedAmounts(claimable_locked=200000, unclaimable_locked=200),
        ),
        ChannelValues(deposit=40, withdrawn=10, transferred=UINT256_MAX - 5),
    ),
    # overflow on transferred amount, overflow on netted transfer + deposit
    (
        ChannelValues(
            deposit=35,
            withdrawn=5,
            transferred=20,
            locked_amounts=LockedAmounts(claimable_locked=200, unclaimable_locked=200000),
        ),
        ChannelValues(deposit=40, withdrawn=10, transferred=UINT256_MAX - 5),
    ),
]



import json
import pprint
import subprocess
import time
from pathlib import Path
from typing import Any, Dict, Optional

import click
import requests
from click import Context
from click.exceptions import BadParameter
from eth_abi import encode_abi

from raiden_contracts.constants import CONTRACT_LIST, CONTRACT_TOKEN_NETWORK, DeploymentModule
from raiden_contracts.contract_manager import (
    ContractManager,
    DeployedContracts,
    contracts_deployed_path,
    contracts_precompiled_path,
    get_contracts_deployment_info,
)
from raiden_contracts.contract_source_manager import (
    contracts_source_path,
    contracts_source_path_of_deployment_module,
)
from raiden_contracts.utils.type_aliases import ChainID

CONTRACT_NAMES_SEPARATED = " | ".join([c.name for c in CONTRACT_LIST])
USER_AGENT = "curl/7.37.0"  # Etherscan blocks us without this user agent in some cases
MODULE_OF_CONTRACT = {contract: module for module, contract in CONTRACT_LIST}


def validate_contract_name(_ctx: Context, _param: Any, value: Optional[str]) -> Optional[str]:
    if value is None:
        return None
    if value in {c.name for c in CONTRACT_LIST}:
        return value
    raise BadParameter(f"unknown contract name {value}")


@click.command()
@click.option("--chain-id", default=3, help="Chain id. E.g. --chain-id 3")
@click.option("--apikey", required=True, help="A valid Etherscan APIKEY is required.")
@click.option(
    "--guid",
    help="GUID from a previous verification attempt. Tries to get the submission status.",
)
@click.option(
    "--contract-name",
    default=None,
    help=f"Contract name. Options: {CONTRACT_NAMES_SEPARATED}. "
    " Default is to submit the sources of all contracts.",
    callback=validate_contract_name,
)
def etherscan_verify(
    chain_id: ChainID, apikey: str, guid: Optional[str], contract_name: Optional[str]
) -> None:
    if guid:
        guid_status(etherscan_api=api_of_chain_id[chain_id], guid=guid, apikey=apikey)
        return

    for list_entry in CONTRACT_LIST:
        if contract_name is None or contract_name == list_entry.name:
            if list_entry.name == CONTRACT_TOKEN_NETWORK:
                _verify_token_networks(chain_id, apikey)
            else:
                _verify_singleton_contract(
                    chain_id=chain_id,
                    apikey=apikey,
                    source_module=list_entry.module,
                    contract_name=list_entry.name,
                )


def _verify_token_networks(chain_id: ChainID, apikey: str) -> None:
    deployment_file_path = contracts_deployed_path(chain_id=chain_id)
    with deployment_file_path.open() as f:
        deployed_contracts_info = json.load(f)
    token_networks = deployed_contracts_info.get("token_networks", [])

    with open(contracts_precompiled_path()) as f:
        contract_dict = json.load(f)["contracts"][CONTRACT_TOKEN_NETWORK]
    metadata = json.loads(contract_dict["metadata"])
    constructor = [func for func in contract_dict["abi"] if func["type"] == "constructor"][0]
    arg_types = [arg["type"] for arg in constructor["inputs"]]
    arg_names = [arg["name"] for arg in constructor["inputs"]]

    for tn in token_networks:
        args = [tn["constructor_arguments"][arg_name] for arg_name in arg_names]
        etherscan_verify_contract(
            chain_id=chain_id,
            apikey=apikey,
            address=tn["token_network_address"],
            contract_name=CONTRACT_TOKEN_NETWORK,
            metadata=metadata,
            constructor_args=encode_abi(arg_types, args).hex(),
        )


api_of_chain_id = {
    1: "https://api.etherscan.io/api",
    3: "https://api-ropsten.etherscan.io/api",
    4: "https://api-rinkeby.etherscan.io/api",
    5: "https://api-goerli.etherscan.io/api",
    42: "https://api-kovan.etherscan.io/api",
}


def join_sources(source_module: DeploymentModule, contract_name: str) -> str:
    """Use join_contracts.py to concatenate all imported Solidity files.

    Args:
        source_module: a module name to look up contracts_source_path()
        contract_name: 'TokenNetworkRegistry', 'SecretRegistry' etc.
    """
    joined_file = Path(__file__).parent.joinpath("joined.sol")
    remapping = {
        module: str(path) for module, path in contracts_source_path(contracts_version=None).items()
    }
    command = [
        "./utils/join_contracts.py",
        "--import-map",
        json.dumps(remapping),
        str(
            contracts_source_path_of_deployment_module(source_module).joinpath(
                contract_name + ".sol"
            )
        ),
        str(joined_file),
    ]
    working_dir = Path(__file__).parent.parent
    try:
        subprocess.check_call(command, cwd=working_dir)
    except subprocess.CalledProcessError as ex:
        print(f"cd {str(working_dir)}; {subprocess.list2cmdline(command)} failed.")
        raise ex

    return joined_file.read_text()


def get_constructor_args(
    deployment_info: DeployedContracts,
    contract_name: str,
    contract_manager: ContractManager,
) -> str:
    constructor_arguments = deployment_info["contracts"][contract_name]["constructor_arguments"]
    if constructor_arguments != []:
        constructor_types = contract_manager.get_constructor_argument_types(contract_name)
        constructor_args = encode_abi(types=constructor_types, args=constructor_arguments).hex()
    else:
        constructor_types = []
        constructor_args = ""
    print("constructor_args", constructor_arguments, constructor_types, constructor_args)
    return constructor_args


def etherscan_verify_contract(
    chain_id: ChainID,
    apikey: str,
    address: str,
    contract_name: str,
    metadata: Dict,
    constructor_args: str,
) -> None:
    source = (
        join_sources(source_module=MODULE_OF_CONTRACT[contract_name], contract_name=contract_name),
    )
    optimization_details = metadata["settings"]["optimizer"].get("details")
    data = {
        # A valid API-Key is required
        "apikey": apikey,
        # Do not change
        "module": "contract",
        # Do not change
        "action": "verifysourcecode",
        "contractaddress": address,
        "sourceCode": source,
        "contractname": contract_name,
        "compilerversion": "v" + metadata["compiler"]["version"],
        # 0 = Optimization used, 1 = No Optimization
        "optimizationUsed": 0 if optimization_details is not None else 1,
        "runs": metadata["settings"]["optimizer"]["runs"],
        # Typo is intentional. Etherscan does not like the correct spelling.
        "constructorArguements": constructor_args,
    }
    pprint.pprint({k: v for k, v in data.items() if k != "sourceCode"})

    etherscan_api = api_of_chain_id[chain_id]
    response = requests.post(etherscan_api, data=data, headers={"User-Agent": USER_AGENT})
    try:
        content = response.json()
    except json.decoder.JSONDecodeError:
        print(response.text)
        raise
    print(content)
    print(f'Status: {content["status"]}; {content["message"]} ; GUID = {content["result"]}')

    etherscan_url = etherscan_api.replace("api-", "").replace("api", "")
    etherscan_url += "/verifyContract2?a=" + data["contractaddress"]
    manual_submission_guide = f"""Usually a manual submission to Etherscan works.
    Visit {etherscan_url}
    Use raiden_contracts/deploy/joined.sol."""

    if content["status"] != "1":
        if content["result"] == "Contract source code already verified":
            return
        else:
            raise ValueError(
                "Etherscan submission failed for an unknown reason\n" + manual_submission_guide
            )

    # submission succeeded, obtained GUID
    guid = content["result"]
    status = "0"
    retries = 10
    while status == "0" and retries > 0:
        retries -= 1
        r = guid_status(etherscan_api=etherscan_api, guid=guid, apikey=apikey)
        status = r["status"]
        if r["result"] == "Fail - Unable to verify":
            raise ValueError(manual_submission_guide)
        if r["result"] == "Pass - Verified":
            return
        print("Retrying...")
        time.sleep(5)
    raise TimeoutError(manual_submission_guide)


def _verify_singleton_contract(
    chain_id: ChainID,
    apikey: str,
    source_module: DeploymentModule,
    contract_name: str,
) -> None:
    """Calls Etherscan API for verifying the Solidity source of a contract.

    This function can only be used to verify contracts which are only deployed
    once. E.g. `TokenNetworkRegistry`, but not `TokenNetwork`.
    Args:
        chain_id: EIP-155 chain id of the Ethereum chain
        apikey: key for calling Etherscan API
        source_module: a module name to look up contracts_source_path()
        contract_name: 'TokenNetworkRegistry', 'SecretRegistry' etc.
    """

    deployment_info = get_contracts_deployment_info(chain_id=chain_id, module=source_module)
    assert deployment_info
    contract_manager = ContractManager(contracts_precompiled_path())
    metadata = json.loads(contract_manager.contracts[contract_name]["metadata"])
    constructor_args = get_constructor_args(
        deployment_info=deployment_info,
        contract_name=contract_name,
        contract_manager=contract_manager,
    )
    contract_deploy_info = deployment_info["contracts"][contract_name]
    etherscan_verify_contract(
        chain_id=chain_id,
        apikey=apikey,
        address=contract_deploy_info["address"],
        contract_name=contract_name,
        metadata=metadata,
        constructor_args=constructor_args,
    )


def guid_status(etherscan_api: str, guid: str, apikey: str) -> Dict:
    data = {
        "guid": guid,
        "module": "contract",
        "action": "checkverifystatus",
        "apikey": apikey,
    }
    r = requests.get(etherscan_api, data, headers={"User-Agent": USER_AGENT})
    try:
        status_content = json.loads(r.content.decode())
    except json.decoder.JSONDecodeError:
        print(r.content.decode())
        raise
    print(status_content)
    return status_content


if __name__ == "__main__":
    # pylint: disable=E1120
    etherscan_verify()
